import pandas as pd
import pytest
import torch

from causalexplain.models import dnn


class DummyTorchModel(torch.nn.Module):
    def __init__(self):
        super().__init__()
        self.loss_fn = torch.nn.MSELoss()

    def forward(self, x):
        return torch.zeros((x.shape[0], 1))


class DummyMLPModel:
    def __init__(self, *args, **kwargs):
        self.args = args
        self.kwargs = kwargs
        self.model = DummyTorchModel()
        self.trained = False

    def train(self):
        self.trained = True


def _small_frame():
    return pd.DataFrame({"a": [0.0, 1.0, 2.0], "b": [1.0, 2.0, 3.0]})


def test_nnregressor_fit_predict_and_score(monkeypatch):
    monkeypatch.setattr(dnn, "MLPModel", DummyMLPModel)
    class FakeProgBar:
        def start_subtask(self, *_, **__):
            return self

        def update_subtask(self, *_, **__):
            return None

        def remove(self, *_, **__):
            return None

    monkeypatch.setattr(dnn, "ProgBar", FakeProgBar)

    df = _small_frame()
    rex = dnn.NNRegressor(
        hidden_dim=[2],
        activation="relu",
        num_epochs=1,
        batch_size=2,
        prog_bar=False,
        early_stop=False,
        min_delta=0.0,
    )
    rex.fit(df)

    preds = rex.predict(df)
    assert preds.shape[0] == len(df.columns)

    scores = rex.score(df)
    assert scores.shape[0] == len(df.columns)
    assert rex.is_fitted_ is True


def test_nnregressor_predict_requires_fit(monkeypatch):
    monkeypatch.setattr(dnn, "MLPModel", DummyMLPModel)
    rex = dnn.NNRegressor()
    with pytest.raises(AttributeError):
        rex.predict(_small_frame())


def test_nnregressor_drops_correlated_features(monkeypatch):
    monkeypatch.setattr(dnn, "MLPModel", DummyMLPModel)

    class FakeHierarchies:
        @staticmethod
        def compute_correlation_matrix(X):
            return X

        @staticmethod
        def compute_correlated_features(matrix, _th, feature_names, verbose=False):
            return {name: [f for f in feature_names if f != name] for name in feature_names}

    monkeypatch.setattr(dnn, "Hierarchies", FakeHierarchies)
    monkeypatch.setattr(dnn, "ProgBar", lambda *_, **__: None)

    df = _small_frame()
    rex = dnn.NNRegressor(correlation_th=0.5, prog_bar=False, early_stop=False, min_delta=0.0, num_epochs=1)
    rex.fit(df)

    # Each model should have received a single predictor after correlated removal.
    for target, model in rex.regressor.items():
        assert model.kwargs["input_size"] == 1


def test_nnregressor_tune_runs_with_fake_optuna(monkeypatch):
    monkeypatch.setattr(dnn, "ColumnsDataset", lambda target, data: data)
    monkeypatch.setattr(
        dnn,
        "DataLoader",
        lambda dataset, batch_size, shuffle=False: [(torch.ones((1, 1)), torch.ones((1, 1)))],
    )

    class DummyModel(torch.nn.Module):
        def __init__(self):
            super().__init__()
            self.loss_fn = torch.nn.MSELoss()

        def forward(self, x):
            return torch.zeros((x.shape[0], 1))

    class DummyWrapper:
        def __init__(self):
            self.model = DummyModel()

    def fake_fit(self, X):
        self.regressor = {name: DummyWrapper() for name in X.columns}
        return self

    monkeypatch.setattr(dnn.NNRegressor, "fit", fake_fit)

    class FakeTrial:
        def __init__(self):
            self.params = {}

        def suggest_int(self, name, low, high):
            self.params[name] = low
            return low

        def suggest_categorical(self, name, options):
            self.params[name] = options[0]
            return options[0]

        def suggest_loguniform(self, name, low, high):
            self.params[name] = low
            return low

        def suggest_uniform(self, name, low, high):
            self.params[name] = low
            return low

        def suggest_float(self, name, low, high):
            self.params[name] = low
            return low

    class FakeFrozen:
        def __init__(self, params, value):
            self.params = params
            self.values = [value]

    class FakeStudy:
        def __init__(self):
            self.best_trials = []
            self.best_value = float("inf")

        def optimize(self, objective, n_trials, show_progress_bar=None, callbacks=None, **kwargs):
            trial = FakeTrial()
            # Run compute_loss with a single repeat to keep things quick.
            original_compute = objective.compute_loss
            objective.compute_loss = lambda model, dataloader, n_repeats=10: original_compute(
                model, dataloader, n_repeats=1
            )
            value = objective(trial)
            self.best_trials = [FakeFrozen(trial.params, value)]
            self.best_value = value

    monkeypatch.setattr(dnn.optuna, "create_study", lambda **kwargs: FakeStudy())
    monkeypatch.setattr(dnn.optuna.logging, "set_verbosity", lambda *_, **__: None)

    df = _small_frame()
    rex = dnn.NNRegressor(num_epochs=1, batch_size=1, prog_bar=False, early_stop=False, min_delta=0.0)
    result = rex.tune(df, df, n_trials=1)
    assert "hidden_dim" in result


def test_nnregressor_tune_fit_and_repr(monkeypatch):
    called = {}

    def fake_tune(self, *_, **__):
        self.min_tunned_loss = 0.25
        return {
            "hidden_dim": [1],
            "activation": "relu",
            "learning_rate": 0.01,
            "dropout": 0.0,
            "batch_size": 1,
            "num_epochs": 1,
        }

    def fake_fit(self, X):
        called["fit"] = True
        self.is_fitted_ = True
        return self

    monkeypatch.setattr(dnn.NNRegressor, "tune", fake_tune)
    monkeypatch.setattr(dnn.NNRegressor, "fit", fake_fit)
    df = _small_frame()

    rex = dnn.NNRegressor()
    rex.tune_fit(df, hpo_n_trials=1)
    assert called.get("fit") is True
    assert "hidden_dim" in repr(rex)


def test_nnregressor_fit_with_progress_and_correlation(monkeypatch):
    monkeypatch.setattr(dnn.utils, "get_feature_names", lambda X: list(X.columns))
    monkeypatch.setattr(
        dnn.utils,
        "get_feature_types",
        lambda X: {list(X.columns)[0]: "categorical", list(X.columns)[1]: "binary"},
    )
    monkeypatch.setattr(dnn, "MLPModel", DummyMLPModel)

    class FakeHierarchies:
        @staticmethod
        def compute_correlation_matrix(X):
            return X

        @staticmethod
        def compute_correlated_features(matrix, _th, feature_names, verbose=False):
            return {name: [f for f in feature_names if f != name] for name in feature_names}

    monkeypatch.setattr(dnn, "Hierarchies", FakeHierarchies)
    monkeypatch.setattr(
        dnn,
        "ProgBar",
        lambda *_, **__: type(
            "PB", (), {"start_subtask": lambda self, *a, **k: self, "update_subtask": lambda *a, **k: None, "remove": lambda *a, **k: None}
        )(),
    )
    monkeypatch.setattr(dnn, "ColumnsDataset", lambda target, data: data)
    monkeypatch.setattr(
        dnn,
        "DataLoader",
        lambda dataset, batch_size, shuffle=False: [
            (torch.ones((len(dataset), 1)), torch.zeros((len(dataset), 1)))
        ],
    )

    df = _small_frame()
    rex = dnn.NNRegressor(
        prog_bar=True, correlation_th=0.5, num_epochs=1, batch_size=1, early_stop=False, min_delta=0.0
    )
    rex.fit(df)
    preds = rex.predict(df)
    scores = rex.score(df)
    assert preds.size > 0
    assert len(scores) == len(df.columns)


def test_nnregressor_fit_handles_stack_failure(monkeypatch):
    def raise_error(*_, **__):
        raise RuntimeError("boom")

    monkeypatch.setattr(dnn.inspect, "getouterframes", raise_error)
    monkeypatch.setattr(dnn, "MLPModel", DummyMLPModel)
    monkeypatch.setattr(dnn.utils, "get_feature_names", lambda X: list(X.columns))
    monkeypatch.setattr(dnn.utils, "get_feature_types", lambda X: {c: "binary" for c in X.columns})
    monkeypatch.setattr(dnn, "ColumnsDataset", lambda target, data: data)
    monkeypatch.setattr(
        dnn,
        "DataLoader",
        lambda dataset, batch_size, shuffle=False: [
            (torch.ones((len(dataset), 1)), torch.zeros((len(dataset), 1)))
        ],
    )
    monkeypatch.setattr(
        dnn,
        "ProgBar",
        lambda *_, **__: type(
            "PB", (), {"start_subtask": lambda self, *a, **k: self, "update_subtask": lambda *a, **k: None, "remove": lambda *a, **k: None}
        )(),
    )

    df = _small_frame()
    rex = dnn.NNRegressor(num_epochs=1, batch_size=1, early_stop=False, min_delta=0.0)
    rex.fit(df)


def test_dnn_custom_main_runs_with_mocks(monkeypatch):
    class DummyScaler:
        def fit_transform(self, data):
            return data.values

    monkeypatch.setattr(dnn.pd, "read_csv", lambda *_: _small_frame())
    monkeypatch.setattr(dnn, "StandardScaler", lambda *_, **__: DummyScaler())
    dnn.custom_main(tune=False, score=False)


def test_dnn_custom_main_tune_branch(monkeypatch):
    class DummyScaler:
        def fit_transform(self, data):
            return data.values

    class DummyNN:
        def __init__(self, *_, **__):
            self.called = False

        def tune_fit(self, *_args, **_kwargs):
            self.called = True

    monkeypatch.setattr(dnn.pd, "read_csv", lambda *_: _small_frame())
    monkeypatch.setattr(dnn, "StandardScaler", lambda *_, **__: DummyScaler())
    monkeypatch.setattr(dnn, "NNRegressor", DummyNN)
    dnn.custom_main(tune=True, score=False)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# causalexplain/__main__.py
#
# (C) 2024 J. Renero
#
# This file is part of causalexplain
#
# pylint: disable=E1101:no-member, W0201:attribute-defined-outside-init
# pylint: disable=C0103:invalid-name, C0116:missing-function-docstring
# pylint: disable=R0913:too-many-arguments, W0511:fixme
# pylint: disable=R0914:too-many-locals, R0915:too-many-statements
# pylint: disable=W0106:expression-not-assigned, R1702:too-many-branches
#

import argparse
import os
import time

import pandas as pd
from causalexplainer import GraphDiscovery

from causalexplain.common import (DEFAULT_BOOTSTRAP_TOLERANCE,
                                  DEFAULT_BOOTSTRAP_TRIALS, DEFAULT_HPO_TRIALS,
                                  DEFAULT_REGRESSORS, DEFAULT_SEED,
                                  HEADER_ASCII, SUPPORTED_METHODS, utils)


def parse_args():
    class SplitArgs(argparse.Action):
        def __call__(self, parser, namespace, values, option_string=None):
            """Split a comma-separated list of values."""
            if values is None:
                setattr(namespace, self.dest, [])
            elif isinstance(values, str):
                setattr(namespace, self.dest, values.split(','))
            else:
                setattr(namespace, self.dest, list(values))

    parser = argparse.ArgumentParser(
        description="Causal Graph Learning with ReX and other compared methods.",
    )
    parser.add_argument(
        '-d', '--dataset', type=str, required=False,
        help='Dataset name. Must be CSV file with headers and comma separated columns')
    parser.add_argument(
        '-m', '--method', type=str, required=False,
        choices=['rex', 'pc', 'fci', 'ges', 'lingam', 'cam', 'notears'],
        help="Method to used. If not specified, the method will be ReX.\n" +
        "Other options are: 'pc', 'fci', 'ges', 'lingam', 'cam', 'notears'.")
    parser.add_argument(
        '-t', '--true_dag', type=str, required=False,
        help='True DAG file name. The file must be in .dot format')
    parser.add_argument(
        '-l', '--load_model', type=str, required=False,
        help='Model name (pickle) to load. If not specified, the model will be trained')
    parser.add_argument(
        '-n', '--no-train', action='store_true', required=False,
        help='Do not train the model. If not specified, the model will be trained')
    parser.add_argument(
        '-T', '--threshold', type=float, required=False,
        help='Threshold to apply to the bootstrapped adjacency matrix.')
    parser.add_argument(
        '-u', '--union', type=str, action=SplitArgs, required=False,
        help='List of comma-separated DAGs from regressor to combine. Default is all of them.')
    parser.add_argument(
        '-i', '--iterations', type=int, required=False,
        help='Hyper-parameter tuning max. iterations')
    parser.add_argument(
        '-b', '--bootstrap', type=int, required=False,
        help='Bootstrap iterations')
    parser.add_argument(
        '-r', '--regressor', type=str, required=False, action=SplitArgs,
        help='Regressor list')
    parser.add_argument(
        '-p', '--prior', type=str, required=False,
        help='Prior file (JSON format) to use in the model')
    parser.add_argument(
        '-S', '--seed', type=int, required=False, help='Random seed')
    parser.add_argument(
        '-s', '--save_model', type=str, required=False, nargs='?', const='',
        help='Save model as specified name. If not specified the model will be saved' +
        ' with the same name as the dataset, but with pickle extension.')
    parser.add_argument(
        '-v', '--verbose', action='store_true', required=False, help='Verbose')
    parser.add_argument(
        '-q', '--quiet', action='store_true', required=False, help='Quiet')
    parser.add_argument(
        '-o', '--output', type=str, required=False,
        help='Output file where saving the final DAG (dot format). If not specified, ' +
        'the final DAG will be printed to stdout.')

    args = parser.parse_args()
    return args


def check_args_validity(args):
    """
    Check the validity of the arguments.

    Returns:
        dict: A dictionary of run values
    """
    run_values = {}

    # Set model type (estimator)
    if args.method is None:
        args.method = 'rex'
        run_values['estimator'] = 'rex'
    else:
        assert args.method in SUPPORTED_METHODS, \
            "Method must be one of: rex, pc, fci, ges, lingam, cam, notears"
        run_values['estimator'] = str(args.method)

    # Check that either dataset is provided or both load_model and no_train
    # are specified
    if args.dataset is None:
        if not (args.load_model and args.no_train):
            raise ValueError(
                "When dataset is not specified, both load_model (-l) and "
                "no_train (-n) options must be provided")
        else:
            run_values['dataset_name'] = None
            run_values['dataset_filepath'] = None
    else:
        # If dataset is provided, check that it exists and load it
        assert os.path.isfile(args.dataset), \
            f"Dataset file '{args.dataset}' does not exist"
        run_values['data'] = pd.read_csv(args.dataset)
        run_values['data_columns'] = list(run_values['data'].columns)
        del run_values['data']
        run_values['dataset_filepath'] = args.dataset

        # Extract the path from where the dataset is, dataset basename
        run_values['dataset_path'] = os.path.dirname(args.dataset)
        dataset_name = os.path.basename(args.dataset)
        run_values['dataset_name'] = dataset_name.replace('.csv', '')

    # Load true DAG, if specified (true_dag)
    run_values['true_dag'] = None
    if args.true_dag is not None:
        assert '.dot' in args.true_dag, "True DAG must be in .dot format"
        assert os.path.isfile(args.true_dag), "True DAG file does not exist"
        run_values['ref_graph'] = utils.graph_from_dot_file(args.true_dag)
        run_values['true_dag'] = args.true_dag

    # Check if 'args.load_mode' does not contain path information. In that case
    # assume it is located in the current directory
    if args.load_model is not None and not os.path.isabs(args.load_model):
        args.load_model = os.path.join(os.getcwd(), args.load_model)
    if args.load_model and not os.path.isfile(args.load_model):
        raise FileNotFoundError("Model file does not exist")
    run_values['load_model'] = args.load_model

    if args.no_train:
        run_values['no_train'] = True
    else:
        run_values['no_train'] = False

    # Load prior, if specified
    run_values['prior'] = None
    if args.prior is not None:
        assert '.json' in args.prior, "Prior file must be in JSON format"
        assert os.path.isfile(args.prior), "Prior file does not exist"
        # Load the JSON file int a List of List of str in run_values['prior']
        run_values['prior'] = utils.read_json_file(args.prior)

    # Determine where to save the model pickle.
    if args.save_model == '':
        save_model = f"{args.dataset.replace('.csv', '')}"
        save_model = f"{save_model}_{args.method}.pickle"
        run_values['save_model'] = os.path.basename(save_model)
        # Output_path is the current directory
        run_values['output_path'] = os.getcwd()
        run_values['model_filename'] = utils.valid_output_name(
            filename=save_model, path=run_values['output_path'])
    elif args.save_model is not None:
        run_values['save_model'] = args.save_model
        run_values['output_path'] = os.path.dirname(args.save_model)
        run_values['model_filename'] = args.save_model
    else:
        run_values['save_model'] = None
        run_values['output_path'] = None

    run_values['seed'] = args.seed if args.seed is not None else DEFAULT_SEED
    run_values['quiet'] = True if args.quiet else False
    run_values['hpo_iterations'] = args.iterations \
        if args.iterations is not None else DEFAULT_HPO_TRIALS
    run_values['bootstrap_iterations'] = args.bootstrap \
        if args.bootstrap is not None else DEFAULT_BOOTSTRAP_TRIALS
    run_values['bootstrap_tolerance'] = args.threshold \
        if args.threshold is not None else DEFAULT_BOOTSTRAP_TOLERANCE

    # Set default regressors in case ReX is called.
    if args.method == 'rex' and args.regressor is None:
        run_values['regressors'] = DEFAULT_REGRESSORS
    else:
        run_values['regressors'] = [args.method]

    run_values['verbose'] = True if args.verbose else False
    run_values['output_dag_file'] = args.output

    # return a dictionary with all the new variables created
    return run_values


def header_():
    """
    Done with "Ogre" font from https://patorjk.com/software/taag/
    """
    print(HEADER_ASCII)


def show_run_values(run_values):
    """
    Print the run values.

    Args:
        run_values (dict): A dictionary of run values
    """
    print("-----")
    print("Run values:")
    for k, v in run_values.items():
        if isinstance(v, pd.DataFrame):
            print(f"- {k}: {v.shape[0]}x{v.shape[1]} DataFrame")
            continue
        print(f"- {k}: {v}")

    print("-----")


def main():
    header_()
    args = parse_args()
    run_values = check_args_validity(args)
    start_time = time.time()

    # Create a new instance of GraphDiscovery
    discoverer = GraphDiscovery(
        experiment_name=run_values['dataset_name'],
        model_type=run_values['estimator'],
        csv_filename=run_values['dataset_filepath'],
        true_dag_filename=run_values['true_dag'],
        verbose=run_values['verbose'],
        seed=run_values['seed']
    )

    if run_values['load_model'] is not None:
        discoverer.load(run_values['load_model'])
        # In case of REX, trainer contains multiple entries in the dictionary
        # and we need to retrieve the last one, but in the case of others, we
        # just need to retrieve the only entry.
        result = next(reversed(discoverer.trainer.values()))
    else:
        discoverer.create_experiments()

    if not run_values['no_train']:
        discoverer.fit_experiments(
            run_values['hpo_iterations'],
            run_values['bootstrap_iterations'],
            run_values['prior']
        )
        result = discoverer.combine_and_evaluate_dags(run_values['prior'])

    elapsed_time, units = utils.format_time(time.time() - start_time)
    print(f"Elapsed time: {elapsed_time:.1f} {units}")
    discoverer.printout_results(result.dag, result.metrics)

    if run_values['output_path'] is not None:
        discoverer.save(run_values['model_filename'])

    if run_values['output_dag_file'] is not None:
        utils.graph_to_dot_file(result.dag, run_values['output_dag_file'])
        print(f"Saved DAG to {run_values['output_dag_file']}")


if __name__ == "__main__":
    main()



import json
from pathlib import Path
from types import SimpleNamespace

import networkx as nx
import numpy as np
import pandas as pd
import pytest

from causalexplain.common import utils


def test_valid_output_name_handles_existing_files(tmp_path):
    existing = tmp_path / "exp.pickle"
    existing.touch()
    candidate = utils.valid_output_name("exp", str(tmp_path), extension="pickle")
    assert candidate.endswith("_1.pickle")
    # Create the suggested name so the next call increments again
    Path(candidate).touch()
    candidate = utils.valid_output_name("exp", str(tmp_path), extension=".pickle")
    assert candidate.endswith("_2.pickle")


def test_save_and_load_experiment_roundtrip(tmp_path):
    obj = {"a": 1}
    path = utils.save_experiment("run", str(tmp_path), obj)
    assert path.endswith(".pickle")
    loaded = utils.load_experiment("run", str(tmp_path))
    assert loaded == obj


def test_graph_from_dictionary_with_weights():
    g = utils.graph_from_dictionary({"u": [("v", 0.3)], "x": ["y"]})
    assert g.has_edge("v", "u") and g["v"]["u"]["weight"] == 0.3
    assert g.has_edge("y", "x")


def test_graph_from_adjacency_threshold_and_inverse():
    adj = np.array([[0, 0.5], [0, 0]])
    g = utils.graph_from_adjacency(adj, node_labels=["a", "b"], th=0.1)
    assert g.has_edge("a", "b")
    g_inv = utils.graph_from_adjacency(adj, inverse=True, absolute_values=True, th=0.1)
    assert g_inv.has_edge(1, 0)


def test_graph_from_adjacency_file_and_to_file(tmp_path):
    mat = np.array([[0, 1], [0, 0]])
    df = pd.DataFrame(mat, columns=["a", "b"])
    adj_file = tmp_path / "adj.csv"
    df.to_csv(adj_file, index=False)
    g, loaded_df = utils.graph_from_adjacency_file(adj_file, th=0.0)
    assert np.array_equal(loaded_df.values, mat)
    assert g.has_edge("a", "b")
    out_file = tmp_path / "out.csv"
    utils.graph_to_adjacency_file(g, out_file, labels=["a", "b"])
    assert out_file.exists()


def test_graph_to_adjacency_symbol_and_weight():
    g = nx.DiGraph()
    g.add_edge("a", "b", **{"b": ">"})
    g.add_edge("b", "a", weight=2)
    mat = utils.graph_to_adjacency(g, labels=["a", "b"])
    assert mat[0, 1] == 2  # symbol ">" maps to 2
    assert mat[1, 0] == 2


def test_select_device_invalid_force():
    with pytest.raises(ValueError):
        utils.select_device(force="baddevice")
    assert utils.select_device(force="cpu") == "cpu"


def test_graph_operations_union_and_intersection():
    g1 = nx.DiGraph()
    g2 = nx.DiGraph()
    g1.add_edge("a", "b")
    g2.add_edge("a", "b")
    nx.set_node_attributes(g1, {"a": {"score": 1}, "b": {"score": 2}})
    nx.set_node_attributes(g2, {"a": {"score": 3}, "b": {"score": 1}})
    inter = utils.graph_intersection(g1, g2)
    assert inter.has_edge("a", "b") and inter.nodes["a"]["score"] == 2
    union = utils.graph_union(g1, g2)
    assert union.has_edge("a", "b")


def test_prior_helpers_and_cycles():
    feature_names = ["x", "y", "z"]
    prior = [["x"], ["y"], ["z"]]
    assert utils.valid_candidates_from_prior(feature_names, "y", prior) == ["x"]
    with pytest.raises(ValueError):
        utils.valid_candidates_from_prior(feature_names, "missing", prior)

    dag = nx.DiGraph([("z", "y"), ("y", "x"), ("x", "z")])
    broken = utils.break_cycles_using_prior(dag, prior, verbose=False)
    assert not list(nx.simple_cycles(broken))


def test_potential_misoriented_edges_and_break_cycles_if_present():
    discrepancies = {
        "a": {"b": SimpleNamespace(shap_gof=0.9)},
        "b": {"a": SimpleNamespace(shap_gof=0.1)},
    }
    loop = ["a", "b"]
    candidates = utils.potential_misoriented_edges(loop, discrepancies)
    assert candidates == [("a", "b", pytest.approx(0.8))]

    dag = nx.DiGraph([("a", "b"), ("b", "a")])
    fixed = utils.break_cycles_if_present(dag, discrepancies)
    assert not list(nx.simple_cycles(fixed))


def test_feature_helpers_and_casting():
    arr = np.array([[0, 1], [1, 2]])
    names = utils.get_feature_names(arr)
    assert names == ["X0", "X1"]
    types = utils.get_feature_types(arr)
    assert types["X0"] == "binary"

    df = pd.DataFrame({"cat": ["a", "b", "a"], "num": [1.0, 2.0, 3.0]})
    casted = utils.cast_categoricals_to_int(df)
    assert casted["cat"].dtype.kind in "iu"


@pytest.mark.parametrize(
    "f1,f2,x_values,expected",
    [
        ([0.1, 0.9], [0.9, 0.1], [0.0, 1.0], 0.5),
        ([0.1, 0.2], [0.3, 0.4], [0.0, 1.0], None),
    ],
)
def test_find_crossing_point(f1, f2, x_values, expected):
    assert utils.find_crossing_point(f1, f2, x_values) == expected


@pytest.mark.parametrize(
    "seconds,expected_unit",
    [(1.0, "s"), (120.0, "m"), (4000.0, "h"), (90000.0, "d"), (700000.0, "w"), (3_000_000.0, "m"), (40_000_000.0, "y"), (400_000_000.0, "a")],
)
def test_format_time_branches(seconds, expected_unit):
    t, unit = utils.format_time(seconds)
    assert unit == expected_unit and t > 0


def test_format_time_negative_raises():
    with pytest.raises(ValueError):
        utils.format_time(-1.0)


def test_stringfy_object_lists_relevant_attributes():
    obj = SimpleNamespace(alpha=1, beta=[1, 2], gamma=np.zeros((1, 1)))
    summary = utils.stringfy_object(obj)
    assert "alpha" in summary and "beta" in summary


def test_combine_dags_and_correct_edges_from_prior():
    dag1 = nx.DiGraph([("a", "b")])
    dag2 = nx.DiGraph([("b", "c")])
    for g in (dag1, dag2):
        g.add_node("a", score=1)
        g.add_node("b", score=1)
        g.add_node("c", score=1)
    discrepancies = {
        "a": {"b": SimpleNamespace(shap_gof=0.2)},
        "b": {"a": SimpleNamespace(shap_gof=0.2), "c": SimpleNamespace(shap_gof=0.3)},
        "c": {"b": SimpleNamespace(shap_gof=0.3)},
    }
    prior = [["a"], ["b"], ["c"]]
    union, inter, u_clean, _ = utils.combine_dags(dag1, dag2, discrepancies, prior=prior)
    assert union.has_edge("a", "b") and union.has_edge("b", "c")
    assert not list(nx.simple_cycles(u_clean))
    assert inter.number_of_edges() == 0


def test_list_files_and_json_reading(tmp_path, monkeypatch):
    (tmp_path / "a.json").write_text("{}", encoding="utf-8")
    (tmp_path / "b.txt").write_text("{}", encoding="utf-8")
    files = utils.list_files("*.json", str(tmp_path))
    assert files == ["a"]
    prior = utils.read_json_file(str(tmp_path / "a.json"))
    assert prior == []
    # Invalid JSON returns empty list
    (tmp_path / "bad.json").write_text("{", encoding="utf-8")
    assert utils.read_json_file(str(tmp_path / "bad.json")) == []
    # Missing file
    assert utils.read_json_file(str(tmp_path / "missing.json")) == []



"""
A class to train DFF networks for all variables in data. Each network will be trained to
predict one of the variables in the data, using the rest as predictors plus one
source of random noise.

(C) 2022,2023,2024, J. Renero
"""

# pylint: disable=E1101:no-member, W0201:attribute-defined-outside-init, W0511:fixme
# pylint: disable=C0103:invalid-name
# pylint: disable=C0116:missing-function-docstring
# pylint: disable=R0913:too-many-arguments
# pylint: disable=R0914:too-many-locals, R0915:too-many-statements
# pylint: disable=W0106:expression-not-assigned, R1702:too-many-branches
# pylint: disable=W0102:dangerous-default-value

import inspect
import types
import warnings
from typing import Dict, List, Tuple, Union, Any

import numpy as np
import optuna
import pandas as pd
import torch
from sklearn.base import BaseEstimator
from sklearn.preprocessing import StandardScaler
from torch.utils.data import DataLoader
from mlforge.progbar import ProgBar   # type: ignore

from ..common import DEFAULT_HPO_TRIALS,  utils
from ..explainability.hierarchies import Hierarchies
from ._columnar import ColumnsDataset
from ._models import MLPModel

warnings.filterwarnings("ignore")


class NNRegressor(BaseEstimator):
    """
    A class to train DFF networks for all variables in data. Each network will be
    trained to predict one of the variables in the data, using the rest as predictors
    plus one source of random noise.

    Attributes:
    -----------
        hidden_dim (int): The dimension(s) of the hidden layer(s). This value
            can be a single integer for DFF or an array with the dimension of
            each hidden layer for the MLP case.
        activation (str): The activation function to use, either 'relu' or 'selu'.
            Default is 'relu'.
        learning_rate (float): The learning rate for the optimizer.
        dropout (float): The dropout rate for the dropout layer.
        batch_size (int): The batch size for the optimizer.
        num_epochs (int): The number of epochs for the optimizer.
        loss_fn (str): The loss function to use. Default is "mmd".
        device (str): The device to use. Either "cpu", "cuda", or "mps". Default
            is "auto".
        test_size (float): The proportion of the data to use for testing. Default
            is 0.1.
        seed (int): The seed for the random number generator. Default is 1234.
        early_stop (bool): Whether to use early stopping. Default is True.
        patience (int): The patience for early stopping. Default is 10.
        min_delta (float): The minimum delta for early stopping. Default is 0.001.
        prog_bar (bool): Whether to enable the progress bar. Default
            is False.

    """

    def __init__(
            self,
            hidden_dim: Union[int, List[int]] = [75, 17],
            activation: str = 'relu',
            learning_rate: float = 0.0046,
            dropout: float = 0.001,
            batch_size: int = 44,
            num_epochs: int = 40,
            loss_fn: str = 'mse',
            device: Union[int, str] = "cpu",
            test_size: float = 0.1,
            early_stop: bool = False,
            patience: int = 10,
            min_delta: float = 0.001,
            correlation_th: float = None,
            random_state: int = 1234,
            verbose: bool = False,
            prog_bar: bool = True,
            silent: bool = False,
            optuna_prog_bar: bool = False):
        """
        Train DFF networks for all variables in data. Each network will be trained to
        predict one of the variables in the data, using the rest as predictors plus one
        source of random noise.

        Args:
            data (pandas.DataFrame): The dataframe with the continuous variables.
            model_type (str): The type of model to use. Either 'dff' or 'mlp'.
            hidden_dim (int): The dimension(s) of the hidden layer(s). This value
                can be a single integer for DFF or an array with the dimension of
                each hidden layer for the MLP case.
            activation (str): The activation function to use, either 'relu' or 'selu'.
                Default is 'relu'.
            learning_rate (float): The learning rate for the optimizer.
            dropout (float): The dropout rate for the dropout layer.
            batch_size (int): The batch size for the optimizer.
            num_epochs (int): The number of epochs for the optimizer.
            loss_fn (str): The loss function to use. Default is "mmd".
            device (str): The device to use. Either "cpu", "cuda", or "mps". Default
                is "auto".
            test_size (float): The proportion of the data to use for testing. Default
                is 0.1.
            seed (int): The seed for the random number generator. Default is 1234.
            early_stop (bool): Whether to use early stopping. Default is True.
            patience (int): The patience for early stopping. Default is 10.
            min_delta (float): The minimum delta for early stopping. Default is 0.001.
            prog_bar (bool): Whether to enable the progress bar. Default
                is False.

        Returns:
            dict: A dictionary with the trained DFF networks, using the name of the
                variables as the key.
        """
        self.hidden_dim = hidden_dim
        self.activation = activation
        self.learning_rate = learning_rate
        self.dropout = dropout
        self.batch_size = batch_size
        self.num_epochs = num_epochs
        self.loss_fn = loss_fn
        self.device = device
        self.test_size = test_size
        self.early_stop = early_stop
        self.patience = patience
        self.min_delta = min_delta
        self.correlation_th = correlation_th
        self.random_state = random_state
        self.verbose = verbose
        self.prog_bar = prog_bar
        self.silent = silent
        self.optuna_prog_bar = optuna_prog_bar

        self.regressor = None
        self._fit_desc = "Training NNs"

        if self.verbose:
            self.prog_bar = False

    def fit(self, X):
        """A reference implementation of a fitting function.

        Parameters
        ----------
        X : {array-like, sparse matrix}, shape (n_samples, n_features)
            The training input samples.
        y : array-like, shape (n_samples,) or (n_samples, n_outputs)
            The target values (class labels in classification, real numbers in
            regression).

        Returns
        -------
        self : object
            Returns self.
        """
        # X, y = check_X_y(X, y)
        self.n_features_in_ = X.shape[1]
        self.feature_names = utils.get_feature_names(X)
        self.feature_types = utils.get_feature_types(X)
        self.regressor = {}

        # Who is calling me?
        try:
            curframe = inspect.currentframe()
            calframe = inspect.getouterframes(curframe, 2)
            caller_name = calframe[1][3]
            if caller_name == "__call__":
                caller_name = "HPO"
        except Exception:                                # pylint: disable=broad-except
            caller_name = "unknown"

        if self.correlation_th:
            self.corr_matrix = Hierarchies.compute_correlation_matrix(X)
            self.correlated_features = Hierarchies.compute_correlated_features(
                self.corr_matrix, self.correlation_th, self.feature_names,
                verbose=self.verbose)
            X_original = X.copy()

        if self.prog_bar and not self.verbose:
            pbar_name = f"({caller_name}) DNN_fit"
            pbar = ProgBar().start_subtask(pbar_name, len(self.feature_names))
        else:
            pbar = None

        for target_idx, target_name in enumerate(self.feature_names):
            # if correlation_th is not None then, remove features that are highly
            # correlated with the target, at each step of the loop
            if self.correlation_th is not None:
                X = X_original.copy()
                if len(self.correlated_features[target_name]) > 0:
                    X = X.drop(self.correlated_features[target_name], axis=1)
                    if self.verbose:
                        print("REMOVED CORRELATED FEATURES: ",
                              self.correlated_features[target_name])

            # Determine Loss function based on the type of target variable
            if self.feature_types[target_name] == 'categorical':
                loss_fn = 'crossentropy'
            elif self.feature_types[target_name] == 'binary':
                loss_fn = 'binary_crossentropy'
            else:
                loss_fn = self.loss_fn

            self.regressor[target_name] = MLPModel(
                target=target_name,
                input_size=X.shape[1],
                activation=self.activation,
                hidden_dim=self.hidden_dim,
                learning_rate=self.learning_rate,
                batch_size=self.batch_size,
                loss_fn=loss_fn,
                dropout=self.dropout,
                num_epochs=self.num_epochs,
                dataframe=X,
                test_size=self.test_size,
                device=self.device,
                seed=self.random_state,
                early_stop=self.early_stop,
                patience=self.patience,
                min_delta=self.min_delta,
                prog_bar=self.prog_bar)
            self.regressor[target_name].train()

            pbar.update_subtask(pbar_name, target_idx+1) if pbar else None

        pbar.remove(pbar_name) if pbar else None
        self.is_fitted_ = True
        return self

    def predict(self, X: pd.DataFrame) -> np.ndarray:
        """Predicts the values for each target variable.

        Parameters
        ----------
        X : pd.DataFrame
            The input data to make predictions on.

        Returns
        -------
        np.ndarray
            The predictions for each target variable.
        """
        assert X.shape[1] == self.n_features_in_, \
            f"X has {X.shape[1]} features, expected {self.n_features_in_}"

        # X = check_array(X, accept_sparse=True)
        if not self.is_fitted_:
            raise ValueError("This Rex instance is not fitted yet. \
                Call 'fit' with appropriate arguments before using this estimator.")

        if self.correlation_th is not None:
            X_original = X.copy()

        prediction = pd.DataFrame(columns=self.feature_names)
        for target in self.feature_names:
            # Remove those features from the loader that are highly correlated to target
            if self.correlation_th is not None:
                X = X_original.drop(self.correlated_features[target], axis=1)

            # Creat a data loader for the target variable. The ColumnsDataset will
            # return the target variable as the second element of the tuple, and
            # will drop the target from the features.
            loader = DataLoader(
                ColumnsDataset(target, X), batch_size=self.batch_size,
                shuffle=False)
            model = self.regressor[target].model

            # Obtain the predictions for the target variable
            preds = np.empty((0,), dtype=np.float16)
            for (tensor_X, _) in loader:
                tensor_X = tensor_X.to(self.device)
                y_hat = model.forward(tensor_X)
                preds = np.append(preds, y_hat.detach().numpy().flatten())
            prediction[target] = preds

        # Concatenate the numpy array for all the batchs
        np_preds = prediction.values
        final_preds = []
        if np_preds.ndim > 1 and np_preds.shape[0] > 1:
            for i in range(len(self.feature_names)):
                column = np_preds[:, i]
                if column.ndim == 1:
                    final_preds.append(column)
                else:
                    final_preds.append(np.concatenate(column))
            final_preds = np.array(final_preds)
        else:
            final_preds = np_preds

        # If final_preds is still 1D, reshape it to 2D
        if final_preds.ndim == 1:
            final_preds = final_preds.reshape(1, -1)

        if len(final_preds) == 0:
            final_preds = np_preds

        return np.array(final_preds)

    def score(self, X):
        """
        Scores the model using the loss function. It returns the list of losses
        for each target variable.
        """
        assert X.shape[1] == self.n_features_in_, \
            f"X has {X.shape[1]} features, expected {self.n_features_in_}"
        if not self.is_fitted_:
            raise ValueError("This Rex instance is not fitted yet. \
                Call 'fit' with appropriate arguments before using this estimator.")

        # Call the class method to predict the values for each target variable
        y_hat = self.predict(X)

        # Handle the case where the prediction returned by the model is not a
        # numpy array but a numpy object type
        if isinstance(y_hat, np.ndarray) and y_hat.dtype == np.object_:
            y_hat = np.vstack(y_hat[:, :].flatten()).astype('float')
            y_hat = torch.Tensor(y_hat)
        scores = []
        for i, target in enumerate(self.feature_names):
            y_preds = torch.Tensor(y_hat[i])
            y = torch.Tensor(X[target].values)
            scores.append(self.regressor[target].model.loss_fn(y_preds, y))

        self.scoring = np.array(scores)
        return self.scoring

    def __repr__(self):
        forbidden_attrs = [
            'fit', 'predict', 'score', 'get_params', 'set_params']
        ret = f"REX object attributes\n"
        ret += f"{'-'*80}\n"
        for attr in dir(self):
            if attr.startswith('_') or \
                attr in forbidden_attrs or \
                    type(getattr(self, attr)) == types.MethodType:
                continue
            elif attr == "X" or attr == "y":
                if isinstance(getattr(self, attr), pd.DataFrame):
                    ret += f"{attr:25} {getattr(self, attr).shape}\n"
                    continue
            elif isinstance(getattr(self, attr), pd.DataFrame):
                ret += f"{attr:25} DataFrame {getattr(self, attr).shape}\n"
            else:
                ret += f"{attr:25} {getattr(self, attr)}\n"

        return ret

    def tune(
            self,
            training_data: pd.DataFrame,
            test_data: pd.DataFrame,
            study_name: str = None,
            min_loss: float = 0.05,
            storage: str = 'sqlite:///rex_tuning.db',
            load_if_exists: bool = True,
            n_trials: int = DEFAULT_HPO_TRIALS) -> Dict[str, Any]:
        """
        Tune the hyperparameters of the model using Optuna.
        """
        class Objective:
            """
            A class to define the objective function for the hyperparameter optimization
            Some of the parameters for NNRegressor have been taken to default values to
            reduce the number of hyperparameters to optimize.

            Include this class in the hyperparameter optimization as follows:

            >>> study = optuna.create_study(direction='minimize',
            >>>                             study_name='study_name_here',
            >>>                             storage='sqlite:///db.sqlite3',
            >>>                             load_if_exists=True)
            >>> study.optimize(Objective(train_data, test_data), n_trials=100)

            The only dependency is you need to pass the train and test data to the class
            constructor. Tha class will build the data loaders for them from the
            dataframes.
            """

            def __init__(
                    self,
                    train_data,
                    test_data,
                    device='cpu',
                    prog_bar=True,
                    verbose=False):
                self.train_data = train_data
                self.test_data = test_data
                self.device = device

                self.n_layers = None
                self.activation = None
                self.learning_rate = None
                self.dropout = None
                self.batch_size = None
                self.num_epochs = None
                self.models = None
                self.prog_bar = prog_bar
                self.verbose = verbose

            def __call__(self, trial):
                """
                This method is called by Optuna to evaluate the objective function.
                """
                self.n_layers = trial.suggest_int('n_layers', 1, 6)
                self.layers = []
                for i in range(self.n_layers):
                    self.layers.append(
                        trial.suggest_int(f'n_units_l{i}', 1, 182))
                self.activation = trial.suggest_categorical(
                    'activation', ['relu', 'selu', 'linear'])
                self.learning_rate = trial.suggest_loguniform(
                    'learning_rate', 1e-5, 1e-1)
                self.dropout = trial.suggest_uniform('dropout', 0.0, 0.5)
                self.batch_size = trial.suggest_int('batch_size', 8, 128)
                self.num_epochs = trial.suggest_int('num_epochs', 10, 80)

                self.models = NNRegressor(
                    hidden_dim=self.layers,
                    activation=self.activation,
                    learning_rate=self.learning_rate,
                    dropout=self.dropout,
                    batch_size=self.batch_size,
                    num_epochs=self.num_epochs,
                    loss_fn='mse',
                    device=self.device,
                    random_state=42,
                    verbose=self.verbose,
                    prog_bar=True & (not self.verbose) & (self.prog_bar),
                    silent=True)

                self.models.fit(self.train_data)

                # Now, measure the performance of the model with the test data.
                loss = []
                for target in list(self.train_data.columns):
                    model = self.models.regressor[target].model
                    loader = DataLoader(
                        ColumnsDataset(target, self.test_data),
                        batch_size=self.batch_size,
                        shuffle=False)
                    avg_loss, _, _ = self.compute_loss(model, loader)
                    loss.append(avg_loss)

                return np.median(loss)

            def compute_loss(
                    self,
                    model: torch.nn.Module,
                    dataloader: torch.utils.data.DataLoader,
                    n_repeats: int = 10) -> Tuple[float, float, np.ndarray]:
                """
                Computes the average MSE loss for a given model and dataloader.

                Parameters:
                -----------
                model: torch.nn.Module
                    The model to compute the loss for.
                dataloader: torch.utils.data.DataLoader
                    The dataloader to use for computing the loss.
                shuffle: int
                    If > 0, the column of the input data to shuffle.

                Returns:
                --------
                avg_loss: float
                    The average MSE loss.
                std_loss: float
                    The standard deviation of the MSE loss.
                losses: np.ndarray
                    The MSE loss for each batch.
                """
                mse = np.array([])
                num_batches = 0
                for _ in range(n_repeats):
                    loss = []
                    for _, (X, y) in enumerate(dataloader):
                        X = X.to(self.device)
                        y = y.to(self.device)
                        yhat = model.forward(X)
                        loss.append(model.loss_fn(yhat, y).item())
                        num_batches += 1
                    if len(mse) == 0:
                        mse = np.array(loss)
                    else:
                        mse = np.vstack((mse, [loss]))

                return np.mean(mse), np.std(mse), mse

        # Callback to stop the study if the MSE is below a threshold.
        def callback(study: optuna.study.Study, trial: optuna.trial.FrozenTrial):
            if trial.value < min_loss or study.best_value < min_loss:
                study.stop()

        if self.verbose is False:
            optuna.logging.set_verbosity(optuna.logging.WARNING)

        # Create and run the HPO study.
        study = optuna.create_study(
            direction='minimize', study_name=study_name, storage=storage,
            load_if_exists=load_if_exists)
        study.optimize(
            Objective(
                training_data, test_data, prog_bar=self.prog_bar, verbose=self.verbose),
            n_trials=n_trials,
            show_progress_bar=(self.optuna_prog_bar & (
                not self.silent) & (not self.verbose)),
            callbacks=[callback]
        )

        # Capture the best parameters and the minimum loss.
        best_trials = sorted(study.best_trials, key=lambda x: x.values[0])
        self.best_params = best_trials[0].params
        self.min_tunned_loss = best_trials[0].values[0]

        if self.verbose and not self.silent:
            print(f"Best params (min loss:{self.min_tunned_loss:.6f}):")
            for k, v in self.best_params.items():
                print(f"\t{k:<15s}: {v}")

        regressor_args = {
            'hidden_dim': [self.best_params[f'n_units_l{i}']
                           for i in range(self.best_params['n_layers'])],
            'activation': self.best_params['activation'],
            'learning_rate': self.best_params['learning_rate'],
            'dropout': self.best_params['dropout'],
            'batch_size': self.best_params['batch_size'],
            'num_epochs': self.best_params['num_epochs']}

        return regressor_args

    def tune_fit(
            self,
            X: pd.DataFrame,
            hpo_study_name: str = None,
            hpo_min_loss: float = 0.05,
            hpo_storage: str = 'sqlite:///rex_tuning.db',
            hpo_load_if_exists: bool = True,
            hpo_n_trials: int = DEFAULT_HPO_TRIALS):
        """
        Tune the hyperparameters of the model using Optuna, and the fit the model
        with the best parameters.
        """
        # split X into train and test
        train_data = X.sample(frac=0.9, random_state=self.random_state)
        test_data = X.drop(train_data.index)

        # tune the model
        regressor_args = self.tune(
            train_data, test_data, n_trials=hpo_n_trials, study_name=hpo_study_name,
            min_loss=hpo_min_loss, storage=hpo_storage,
            load_if_exists=hpo_load_if_exists)

        if self.verbose and not self.silent:
            print(f"Best params (min loss:{self.min_tunned_loss:.6f}):")
            for k, v in regressor_args.items():
                print(f"\t{k:<15s}: {v}")

        # Set the object parameters to the best parameters found.
        for k, v in regressor_args.items():
            setattr(self, k, v)

        # Fit the model with the best parameters.
        self.fit(X)


#
# Main function
#

def custom_main(score: bool = False, tune: bool = False):
    import os
    from causalexplain.common import utils
    path = "/Users/renero/phd/data/RC4/risks"
    output_path = "/Users/renero/phd/output/RC4/"
    experiment_name = 'transformed_data'

    # ref_graph = utils.graph_from_dot_file(f"{path}{experiment_name}.dot")

    data = pd.read_csv(f"{os.path.join(path, experiment_name)}.csv")
    scaler = StandardScaler()
    data = pd.DataFrame(scaler.fit_transform(data), columns=data.columns)

    # Split the dataframe into train and test
    train = data.sample(frac=0.9, random_state=42)
    test = data.drop(train.index)

    if score:
        rex = utils.load_experiment(f"{experiment_name}", output_path)
        rex.is_fitted_ = True
        print(f"Loaded experiment {experiment_name}")
        rex.models.score(test)
    elif tune:
        nn = NNRegressor()
        nn.tune_fit(data, hpo_n_trials=10)


if __name__ == "__main__":
    custom_main(tune=True)



import networkx as nx
from causalexplain.independence.cond_indep import (
    ConditionalIndependencies, get_backdoor_paths, get_paths,
    find_colliders_in_path, get_sufficient_sets_for_pair
)

# Returns an empty list if x and y are not connected


def test_returns_empty_list_if_not_connected():
    """
    Test that the function get_backdoor_paths returns an empty list when there 
    is no path between x and y in the DAG.
    """
    # Create a simple DAG
    dag = nx.DiGraph()
    dag.add_edges_from([(1, 2), (2, 3), (2, 4), (3, 4), (4, 5), (5, 6)])

    # Test case: x and y are not connected
    x = 1
    y = 6
    paths = get_backdoor_paths(dag, x, y)
    assert paths == []

# Returns an empty list if there are no backdoor paths between x and y


def test_returns_empty_list_if_no_backdoor_paths():
    """
    Test that the function get_backdoor_paths returns an empty list when 
    there are no backdoor paths between two nodes.
    """
    # Create a simple DAG
    dag = nx.DiGraph()
    dag.add_edges_from([(1, 2), (2, 3), (2, 4), (3, 4), (4, 5), (5, 6)])

    # Test case: x and y are connected, but there are no backdoor paths
    x = 1
    y = 2
    paths = get_backdoor_paths(dag, x, y)
    assert paths == []

# Returns a list with one path if there is one backdoor path between x and y


def test_returns_list_with_one_path_if_one_backdoor_path():
    """
    Test that the function get_backdoor_paths returns a list with one path if 
    there is only one backdoor path between two nodes.
    """
    # Create a simple DAG
    dag = nx.DiGraph()
    dag.add_edges_from([(2, 1), (2, 3), (2, 4), (3, 4), (4, 5), (5, 6)])

    # Test case: x and y are connected, and there is one backdoor path
    x = 1
    y = 4
    paths = get_backdoor_paths(dag, x, y)
    assert set(map(tuple, paths)) == {tuple([1, 2, 4]), tuple([1, 2, 3, 4])}

# Returns an empty list if x and y are the same node


def test_returns_condsets_when_collider():
    """
    Test function to check if the function get_backdoor_paths returns the 
    correct set of paths
    when there is a collider in the graph.
    """
    dag = nx.DiGraph()
    dag.add_edges_from([('r', 'x'), ('r', 's'), ('t', 's'), ('t', 'y')])

    x = 'x'
    y = 'y'
    paths = get_backdoor_paths(dag, x, y)
    print(paths)
    assert set(map(tuple, paths)) == {tuple(['x', 'r', 's', 't', 'y'])}


def test_returns_empty_list_if_same_node():
    """
    Test that the function get_backdoor_paths returns an empty list when the 
    input nodes are the same.

    The function should return an empty list because there are no backdoor paths 
    between a node and itself.
    """
    # Create a simple DAG
    dag = nx.DiGraph()
    dag.add_edges_from([(1, 2), (2, 3), (2, 4), (3, 4), (4, 5), (5, 6)])

    # Test case: x and y are the same node
    x = 1
    y = 1
    paths = get_backdoor_paths(dag, x, y)
    assert paths == []

# Returns an empty list if x and y are not in the graph


def test_returns_empty_list_if_not_in_graph():
    """
    Test that the function get_backdoor_paths returns an empty list when 
    x and y are not in the graph.
    """
    # Create a simple DAG
    dag = nx.DiGraph()
    dag.add_edges_from([(1, 2), (2, 3), (2, 4), (3, 4), (4, 5), (5, 6)])

    # Test case: x and y are not in the graph
    x = 7
    y = 8
    paths = get_backdoor_paths(dag, x, y)
    assert paths == []

# Returns an empty list if there is no path between x and y


def test_returns_empty_list_if_no_path():
    """
    Test that the function get_backdoor_paths returns an empty list when 
    there is no path between x and y in the DAG.
    """
    # Create a simple DAG
    dag = nx.DiGraph()
    dag.add_edges_from([(1, 2), (2, 3), (2, 4), (3, 4), (4, 5), (5, 6)])

    # Test case: there is no path between x and y
    x = 1
    y = 6
    paths = get_backdoor_paths(dag, x, y)
    assert paths == []


def test_get_paths_returns_empty_list_if_not_connected():
    """
    Test that the function get_paths returns an empty list when there 
    is no path between x and y in the DAG.
    """
    # Create a simple DAG
    dag = nx.DiGraph()
    dag.add_edges_from([(2, 3), (2, 4), (3, 4), (4, 5), (5, 6)])

    # Test case: x and y are not connected
    x = 1
    y = 6
    paths = get_paths(dag, x, y)
    assert paths == []


def test_get_paths_returns_list_with_one_path_if_one_path():
    """
    Test that the function get_paths returns a list with one path if 
    there is only one path between two nodes.
    """
    # Create a simple DAG
    dag = nx.DiGraph()
    dag.add_edges_from([(1, 2), (2, 3), (3, 4), (4, 5), (5, 6)])

    # Test case: x and y are connected, and there is one path
    x = 1
    y = 6
    paths = get_paths(dag, x, y)
    assert paths == [[1, 2, 3, 4, 5, 6]]


def test_get_paths_returns_list_with_multiple_paths():
    """
    Test that the function get_paths returns a list with multiple paths if 
    there are multiple paths between two nodes.
    """
    # Create a simple DAG
    dag = nx.DiGraph()
    dag.add_edges_from([(1, 2), (2, 3), (2, 4), (3, 4), (4, 5), (5, 6)])

    # Test case: x and y are connected, and there are multiple paths
    x = 1
    y = 6
    paths = get_paths(dag, x, y)
    assert set(map(tuple, paths)) == {
        tuple([1, 2, 4, 5, 6]), tuple([1, 2, 3, 4, 5, 6])}


def test_get_paths_returns_empty_list_if_same_node():
    """
    Test that the function get_paths returns an empty list when the 
    input nodes are the same.

    The function should return an empty list because there are no paths 
    between a node and itself.
    """
    # Create a simple DAG
    dag = nx.DiGraph()
    dag.add_edges_from([(1, 2), (2, 3), (2, 4), (3, 4), (4, 5), (5, 6)])

    # Test case: x and y are the same node
    x = 1
    y = 1
    paths = get_paths(dag, x, y)
    assert paths == []


def test_get_paths_returns_empty_list_if_not_in_graph():
    """
    Test that the function get_paths returns an empty list when 
    x and y are not in the graph.
    """
    # Create a simple DAG
    dag = nx.DiGraph()
    dag.add_edges_from([(1, 2), (2, 3), (2, 4), (3, 4), (4, 5), (5, 6)])

    # Test case: x and y are not in the graph
    x = 7
    y = 8
    paths = get_paths(dag, x, y)
    assert paths == []


def test_get_paths_returns_empty_list_if_no_path():
    """
    Test that the function get_paths returns an empty list when 
    there is no path between x and y in the DAG.
    """
    # Create a simple DAG
    dag = nx.DiGraph()
    dag.add_edges_from([(1, 2), (2, 3), (2, 4), (3, 4), (4, 5), (7, 6)])

    # Test case: there is no path between x and y
    x = 1
    y = 6
    paths = get_paths(dag, x, y)
    assert paths == []


def test_find_colliders_in_path_returns_empty_set_when_no_colliders():
    """
    Test that the function find_colliders_in_path returns an empty set when 
    there are no colliders in the path.
    """
    # Create a simple DAG
    dag = nx.DiGraph()
    dag.add_edges_from([(1, 2), (2, 3), (3, 4), (4, 5), (5, 6)])

    # Test case: there are no colliders in the path
    path = [1, 2, 3, 4, 5, 6]
    colliders = find_colliders_in_path(dag, path)
    assert colliders == set()


def test_find_colliders_in_path_returns_set_of_colliders():
    """
    Test that the function find_colliders_in_path returns a set of colliders 
    in the path.
    """
    # Create a simple DAG
    dag = nx.DiGraph()
    dag.add_edges_from([(1, 2), (2, 3), (3, 4), (5, 4), (5, 6)])

    # Test case: there are colliders in the path
    path = [1, 2, 3, 4, 5, 6]
    colliders = find_colliders_in_path(dag, path)
    assert colliders == {4}


def test_find_colliders_in_path_returns_empty_set_when_path_is_too_short():
    """
    Test that the function find_colliders_in_path returns an empty set when 
    the path is too short to contain a collider.
    """
    # Create a simple DAG
    dag = nx.DiGraph()
    dag.add_edges_from([(1, 2), (2, 3), (3, 4), (4, 5), (5, 6)])

    # Test case: the path is too short to contain a collider
    path = [1, 2]
    colliders = find_colliders_in_path(dag, path)
    assert colliders == set()


def test_find_colliders_in_path_returns_empty_set_when_path_is_not_in_dag():
    """
    Test that the function find_colliders_in_path returns an empty set when 
    the path is not in the DAG.
    """
    # Create a simple DAG
    dag = nx.DiGraph()
    dag.add_edges_from([(1, 2), (2, 3), (3, 4), (4, 5), (5, 6)])

    # Test case: the path is not in the DAG
    path = [1, 2, 3, 7]
    colliders = find_colliders_in_path(dag, path)
    assert colliders == set()


def test_add_new_independence():
    ci = ConditionalIndependencies()
    ci.add('X', 'Y', ['Z'])
    assert ci._cache[('X', 'Y', ('Z',))] is True


def test_add_existing_independence():
    ci = ConditionalIndependencies()
    ci.add('X', 'Y', ['Z'])
    ci.add('X', 'Y', ['Z'])  # Adding the same independence again
    assert len(ci._cache) == 1  # Cache size should remain 1


def test_add_empty_conditioning_set():
    ci = ConditionalIndependencies()
    ci.add('A', 'B', [])
    assert ci._cache[('A', 'B', ())] is True


def test_no_backdoor_paths():
    dag = nx.DiGraph()
    dag.add_edges_from([('A', 'B'), ('C', 'D')])
    sufficient_sets = get_sufficient_sets_for_pair(dag, 'A', 'D')
    assert sufficient_sets == []


def test_single_backdoor_path():
    dag = nx.DiGraph()
    dag.add_edges_from([('Z', 'X'), ('Z', 'Y')])
    sufficient_sets = get_sufficient_sets_for_pair(dag, 'X', 'Y')
    assert sufficient_sets == [['Z']]


def test_path_with_descendants():
    dag = nx.DiGraph()
    dag.add_edges_from([('X', 'Z'), ('Z', 'Y'), ('X', 'Z2')])
    sufficient_sets = get_sufficient_sets_for_pair(dag, 'X', 'Y')
    assert sufficient_sets == []


def test_path_with_colliders():
    dag = nx.DiGraph()
    dag.add_edges_from([('X', 'Z'), ('Z', 'Y'), ('Z', 'X')])
    sufficient_sets = get_sufficient_sets_for_pair(dag, 'X', 'Y')
    assert sufficient_sets == []



import pandas as pd
import torch

from causalexplain.models import _models
from causalexplain.models._models import BaseModel, MLPModel


class DummyLogger:
    def __init__(self, *args, **kwargs):
        self.args = args
        self.kwargs = kwargs


class DummyTrainer:
    def __init__(self, *args, **kwargs):
        self.args = args
        self.kwargs = kwargs
        self.logger = kwargs.get("logger")
        self.fit_called = False

    def fit(self, *_, **__):
        self.fit_called = True
        return None


def _small_dataframe():
    return pd.DataFrame(
        {
            "target": [0.0, 1.0, 2.0, 3.0, 4.0],
            "a": [1, 2, 3, 4, 5],
            "b": [5, 4, 3, 2, 1],
        }
    )


def test_base_model_initializes_loaders_and_batch(monkeypatch):
    monkeypatch.setattr(_models, "TensorBoardLogger", DummyLogger)
    df = _small_dataframe()
    base = BaseModel(
        target="target",
        dataframe=df,
        test_size=0.6,
        batch_size=4,
        tb_suffix="test",
        seed=0,
        early_stop=False,
    )

    # Validation split was smaller than requested batch size, so it shrinks.
    assert base.batch_size == 3
    assert len(base.train_loader.dataset) + len(base.val_loader.dataset) == len(df)


def test_override_extras_allows_custom_values(monkeypatch):
    monkeypatch.setattr(_models, "TensorBoardLogger", DummyLogger)
    df = _small_dataframe()
    base = BaseModel(
        target="target",
        dataframe=df,
        test_size=0.5,
        batch_size=2,
        tb_suffix="test",
        seed=0,
        early_stop=False,
    )
    base.n_rows = 12
    base.batch_size = 3

    base.override_extras(log_every_n_steps=7, fast_dev_run=True)
    assert base.extra_trainer_args["log_every_n_steps"] == 7
    assert base.extra_trainer_args["fast_dev_run"] is True
    assert base.extra_trainer_args["enable_model_summary"] is False


def test_mlp_model_initializes_trainer(monkeypatch):
    monkeypatch.setattr(_models, "TensorBoardLogger", DummyLogger)
    monkeypatch.setattr(_models, "Trainer", DummyTrainer)
    df = _small_dataframe()

    model = MLPModel(
        target="target",
        input_size=df.shape[1],
        hidden_dim=[2],
        activation="relu",
        learning_rate=0.01,
        batch_size=2,
        loss_fn="mse",
        dropout=0.0,
        num_epochs=1,
        dataframe=df,
        test_size=0.4,
        device="cpu",
        seed=0,
        early_stop=False,
        log_every_n_steps=5,
    )

    assert isinstance(model.trainer, DummyTrainer)
    assert model.trainer.kwargs["accelerator"] == "cpu"
    assert model.trainer.kwargs["log_every_n_steps"] == 5


def test_init_callbacks_adds_progress_bar(monkeypatch):
    monkeypatch.setattr(_models, "TensorBoardLogger", DummyLogger)

    class DummyBar:
        def init_train_tqdm(self):
            return type("TQDM", (), {"dynamic_ncols": False, "ncols": 0})()

    monkeypatch.setattr(_models, "TQDMProgressBar", DummyBar)
    df = _small_dataframe()
    base = BaseModel(
        target="target",
        dataframe=df,
        test_size=0.5,
        batch_size=2,
        tb_suffix="test",
        seed=0,
    )
    base.init_callbacks(early_stop=True, prog_bar=True)
    bar = next(cb for cb in base.callbacks if isinstance(cb, DummyBar))
    result = bar.init_train_tqdm()
    assert getattr(result, "ncols", None) == 0
    assert any(cb.__class__.__name__ == "EarlyStopping" for cb in base.callbacks)


def test_mlp_model_train_invokes_trainer(monkeypatch):
    monkeypatch.setattr(_models, "TensorBoardLogger", DummyLogger)
    dummy_trainer = DummyTrainer()
    monkeypatch.setattr(_models, "Trainer", lambda *a, **k: dummy_trainer)
    df = _small_dataframe()

    model = MLPModel(
        target="target",
        input_size=df.shape[1],
        hidden_dim=[2],
        activation="relu",
        learning_rate=0.01,
        batch_size=2,
        loss_fn="mse",
        dropout=0.0,
        num_epochs=1,
        dataframe=df,
        test_size=0.4,
        device="cpu",
        seed=0,
    )

    model.train()
    assert dummy_trainer.fit_called is True



import networkx as nx
import numpy as np
import pandas as pd
import pytest

import causalexplain.independence.graph_independence as gi_mod
from causalexplain.independence.graph_independence import GraphIndependence


class StubHsic:
    def __init__(self, pvalue):
        self.pvalue = pvalue
        self.calls = []

    def test(self, x, y):
        self.calls.append((x, y))
        return 0.0, self.pvalue


def _base_instance(condsize=1):
    base_graph = nx.DiGraph([("x", "y"), ("x", "z")])
    gi = GraphIndependence(
        base_graph, condlen=1, condsize=condsize, prog_bar=False, verbose=True
    )
    gi.G_skl = base_graph.copy()
    gi.data = pd.DataFrame({"x": [0, 1], "y": [1, 2], "z": [2, 3]})
    gi.sepset = {("x", "y"): (), ("y", "x"): ()}
    return gi


def test_gen_cond_sets_and_empty_helper():
    gi = _base_instance(condsize=0)
    assert gi._gen_cond_sets("x", "y", 0) == [()]

    gi.G_skl = nx.DiGraph([("x", "y"), ("x", "z"), ("y", "z")])
    conds = gi._gen_cond_sets("x", "y", 1)
    assert conds == [("z",)]
    assert gi._empty([]) and not gi._empty([("z",)])


def test_cond_indep_test_branches(monkeypatch):
    gi = _base_instance()

    # No conditioning set
    monkeypatch.setattr(gi_mod, "Hsic", lambda: StubHsic(0.51))
    indep, _, pval = gi._cond_indep_test(
        gi.data["x"].values.reshape(-1, 1),
        gi.data["y"].values.reshape(-1, 1),
        None,
    )
    assert indep and pval == 0.51

    # With conditioning set goes through residuals branch
    monkeypatch.setattr(gi_mod, "Hsic", lambda: StubHsic(0.01))
    monkeypatch.setattr(gi, "_residuals", lambda *args: (np.array([1]), np.array([2])))
    indep, _, pval = gi._cond_indep_test(
        gi.data["x"].values.reshape(-1, 1),
        gi.data["y"].values.reshape(-1, 1),
        gi.data[["z"]].values,
    )
    assert not indep and pval == 0.01


def test_test_cond_independence_updates_graph(monkeypatch):
    gi = _base_instance()
    gi.G_skl.add_edge("y", "x")

    monkeypatch.setattr(
        gi, "_cond_indep_test", lambda *args, **kwargs: (True, 0.0, 0.2)
    )
    result = gi._test_cond_independence("x", "y", [("z",), ()])

    assert result == ("x", "y", ("z",))
    assert not gi.G_skl.has_edge("x", "y")
    assert gi.sepset[("x", "y")] == ("z",)


def test_remove_independent_edges_records_actions(monkeypatch):
    gi = _base_instance()
    gi.actions = None
    monkeypatch.setattr(gi, "_gen_cond_sets", lambda *_, **__: [("z",)])
    monkeypatch.setattr(
        gi,
        "_test_cond_independence",
        lambda x, y, cond: ("x", y, ("z",)) if y == "y" else None,
    )

    gi._remove_independent_edges("x", condlen=0, condsize=1)
    assert gi.actions["x"] == [("y", ("z",))]


def test_compute_cond_indep_pvals(monkeypatch):
    gi = _base_instance()
    gi.feature_names = list(gi.data.columns)
    monkeypatch.setattr(gi, "_cond_indep_test", lambda *args, **kwargs: (False, 0.0, 0.3))

    pvals = gi.compute_cond_indep_pvals()
    assert len(pvals) == 6  # 3 features -> 3*2 ordered pairs
    assert all(val == 0.3 for val in pvals.values())



# pylint: disable=E1101:no-member, W0201:attribute-defined-outside-init, W0511:fixme
# pylint: disable=C0103:invalid-name
# pylint: disable=C0116:missing-function-docstring
# pylint: disable=R0913:too-many-arguments
# pylint: disable=R0914:too-many-locals, R0915:too-many-statements
# pylint: disable=W0106:expression-not-assigned, R1702:too-many-branches

from copy import copy
from typing import List

import colorama
import numpy as np
import pandas as pd
from sklearn import metrics
from sklearn.cluster import DBSCAN
from sklearn.metrics import euclidean_distances

BLACK = colorama.Fore.BLACK
RED = colorama.Fore.RED
GREEN = colorama.Fore.GREEN
GRAY = colorama.Fore.LIGHTBLACK_EX
RESET = colorama.Style.RESET_ALL


def select_features(
        values,
        feature_names,
        return_shaps=False,
        min_impact: float = 1e-06,
        exhaustive=False,
        threshold: float = None,
        verbose=False) -> List[str]:
    """
    Sort the values and select those before (strict) the point of max. curvature,
    according to the selected algorithm. If strict is False, the point of max curv.
    is also selected. When the method is 'abrupt' the selection method is based on
    taking only those feature up-to (down-to) a certain percentage of change in their
    values.

    Arguments:
        - values (np.ndarray): The values for each of the features. This can
            be anything that should be used to determine what features are more
            important than others.
        - feature_names (list): Names of the variables corresponding to the shap values
        - return_shaps (bool): Whether returning the mean shap values together with
            order of the features.
        - min_impact (float): Default 1e-06. The minimum impact of a feature to be
            selected. If all features are below this value, none are selected.
        - exhaustive (bool): Default False. Whether to use the exhaustive method or
            not. If True, the threshold is used to find all possible clusters above
            the given threshold, not only the first one.
        - threshold (float): Default None. The threshold to use when exahustive is
            True. If None, exception is raised.
        - verbose: guess what.

    """
    if (exhaustive is True) and (threshold is None):
        raise ValueError("If exhaustive is True, threshold must be provided.")
    threshold = 0.0 if threshold is None else threshold

    if len(values.shape) > 1:
        feature_order = np.argsort(np.sum(np.abs(values), axis=0))
        mean_values = np.abs(values).mean(0)
    else:
        feature_order = np.argsort(np.abs(values))
        mean_values = np.abs(values)
    sorted_shap_values = np.array([mean_values[idx] for idx in feature_order])
    if verbose:
        print(f"  Feature order......: {feature_order}")

    # In some cases, the mean SHAP values are 0. We return an empty list in that case.
    if np.all(mean_values < min_impact):
        return []

    if verbose:
        print("  Sum values.........: ", end="")
        if len(values.shape) > 1:
            print(','.join([f"({f}:{s:.03f})" for f, s in zip(
                feature_names, np.sum(np.abs(values), axis=0))]))
        else:
            print(','.join([f"({f}:{s:.03f})" for f, s in zip(
                feature_names, np.abs(values))]))
        print(
            f"  Feature_order......: "
            f"{','.join([f'{feature_names[i]}' for i in feature_order])}\n"
            f"  sorted_mean_values.: "
            f"{','.join([f'{x:.6f}' for x in sorted_shap_values])}\n"
            f"  threshold..........: {threshold:.6f}")

    sorted_impact_values = copy(sorted_shap_values)
    selected_features = []
    max_iterations = len(sorted_impact_values)
    iteration = 0
    limit_idx = 0
    while np.any(sorted_impact_values > threshold):
        if iteration >= max_iterations:
            break

        limit_idx = find_cluster_change_point(sorted_impact_values, verbose=verbose)
        selected_features = list(reversed(
            [feature_names[i] for i in feature_order[limit_idx:]]))

        if not exhaustive:
            break

        sorted_impact_values = sorted_impact_values[:limit_idx]
        iteration += 1

    if verbose:
        print(f"  Limit_idx(cut-off).: {limit_idx}")
        print(f"  Selected_features..: {selected_features}")
    if return_shaps:
        return selected_features, list(reversed(sorted(mean_values)[limit_idx:]))

    return selected_features


def find_cluster_change_point(X: List, verbose: bool = False) -> int:
    """
    Given an array of values in increasing or decreasing order, detect what are the
    elements that belong to the same cluster. The clustering is done using DBSCAN
    with a distance computed as the max. difference between consecutive elements.

    Arguments:
        - X (np.array): the series of values to detect the abrupt change.
        - verbose (bool): Verbose output.

    Returns:
        The position in the array where an abrupt change is produced. If there's
            no change in consecutive values greater than the tolerance passed then
            the last element of the array is returned.
    """
    if len(X) <= 1:
        return None

    X = np.array(X).reshape(-1, 1)
    X_safe = X.copy()
    X_safe[X_safe == 0.0] = 1e-06

    pairwise_distances = euclidean_distances(X)[0,]
    pairwise_distances = np.diff(pairwise_distances)
    pairwise_distances = np.sort(pairwise_distances)[::-1]

    # Safety check: if any of the values in pairwise_distances is 0, add 1e-06
    pairwise_distances[pairwise_distances == 0.0] = 1e-06

    n_clusters_ = 0
    while n_clusters_ <= 1 and len(pairwise_distances) > 0:
        max_distance = pairwise_distances.max() + 1e-06

        # Safety check
        max_distance = 0.001 if max_distance <= 0.0 else max_distance

        if verbose:
            print(f"  pairwise_distances.: {pairwise_distances}")
            print(f"  max_distance.......: {max_distance:.4f}")

        db = DBSCAN(eps=max_distance, min_samples=1).fit(X)
        labels = db.labels_
        # Number of clusters in labels, ignoring noise if present.
        n_clusters_ = len(set(labels)) - (1 if -1 in labels else 0)
        n_noise_ = list(labels).count(-1)
        if n_clusters_ <= 1:
            if verbose:
                print("    ↳ Only 1 cluster generated. Decreasing max_distance.")
            pairwise_distances = pairwise_distances[1:]

    if pairwise_distances.size == 0:
        print("** No clusters generated") if verbose else None
        return None

    if verbose:
        print(f"  Est.clusters/noise.: {n_clusters_}/{n_noise_}")
        if (len(labels) > 3) and (len(labels) < (X.shape[0]-1)):
            print(
                f"  Silhouette Coeff...: {metrics.silhouette_score(X, labels):.3f}\n"
                f"    ↳ Labels: {labels}")

    winner_label = n_clusters_ - 1
    samples_in_winner_cluster = np.argwhere(X_safe[labels != winner_label])

    return samples_in_winner_cluster[:, 0][-1]+1


def main():
    # Display Options
    np.set_printoptions(precision=4, linewidth=100)
    pd.set_option('display.precision', 4)
    pd.set_option('display.float_format', '{:.4f}'.format)

    toy_values = np.array(
        [0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 1.959290610692456]
    )
    # values = np.array(
    #     [0.083, 0.175, 0.353, 0.204, 0.081, 0.116, 0.088, 0.451, 0.152]
    # )
    names = [f"V{i}" for i in range(len(toy_values))]
    select_features(toy_values, names, verbose=True)


def test():
    values = np.array([0.1, 0.2, 0.3, 0.4, 0.5])
    feature_names = ['A', 'B', 'C', 'D', 'E']
    result = select_features(values, feature_names)
    assert result == ['A', 'B', 'C', 'D', 'E']


if __name__ == "__main__":
    test()



"""

- The `pruning` function is translated to Python.
- `dim(G)[1]` is replaced with `G.shape[0]` to get the number of rows.
- `matrix(0,p,p)` is replaced with `np.zeros((p, p))` to create a zero matrix.
- `which(G[,i]==1)` is replaced with `np.where(G[:, i] == 1)[0]` to find the indices
    where the condition is true.
- `cbind(X[,parents],X[,i])` is replaced with `np.hstack((X[:, parents], X[:, [i]]))`
    to concatenate arrays horizontally.
- The `cat` function is replaced with `print` for output.
- The `pruneMethod` function is passed as `prune_method` and called accordingly.
"""
import numpy as np

from .selGam import selGam


def pruning(
        X,
        G,
        verbose=False,
        prune_method=None,
        prune_method_pars={'cutOffPVal': 0.001, 'numBasisFcts': 10}):
    """_summary_

    Args:
        X (_type_): Input vectors
        G (_type_): Adjacency matrix representing a DAG
        output (bool, optional): Whether to print debug messages
        prune_method (_type_, optional): _description_. Defaults to None.
        prune_method_pars (dict, optional): _description_.
            Defaults to {'cutOffPVal': 0.001, 'numBasisFcts': 10}.

    Returns:
        _type_: _description_
    """
    if prune_method is None:
        prune_method = selGam

    p = G.shape[0]
    finalG = np.zeros((p, p))

    for i in range(p):
        parents = np.where(G[:, i] == 1)[0]
        lenpa = len(parents)

        if verbose:
            print(f"Pruning variable: {i}")
            print(f". Considered parents: {parents}")

        if lenpa > 0:
            Xtmp = np.hstack((X[:, parents], X[:, [i]]))
            selected_par = prune_method(
                Xtmp, pars=prune_method_pars, verbose=verbose, k=lenpa+1)
            final_parents = parents[selected_par]
            finalG[final_parents, i] = 1
            if verbose:
                print(f". Final parents: {final_parents}")
                print(f". Pruned parents of {i}: {[p for p in parents if p not in final_parents]}")

    return finalG



"""
- **Function Definition**: Translated the R function `computeScoreMat` to Python as
`compute_score_mat`.
- **Combinations**: Used `itertools.combinations` to generate parent combinations.
- **DataFrame**: Used `pandas.DataFrame` to create the grid of indices.
- **Parallel Processing**: Used `multiprocessing.Pool` for parallel processing.
- **Variance Calculation**: Used `numpy.var` to calculate variance and adjusted the
score matrix accordingly.

Make sure the `compute_score_mat_parallel` function is correctly defined in
`computeScoreMatParallel.py`. If you need further assistance with that function,
please provide its R code.
"""

# pylint: disable=E1101:no-member, W0201:attribute-defined-outside-init, W0511:fixme
# pylint: disable=C0103:invalid-name, W0221:arguments-differ
# pylint: disable=C0116:missing-function-docstring
# pylint: disable=R0913:too-many-arguments, E0401:import-error
# pylint: disable=R0914:too-many-locals, R0915:too-many-statements
# pylint: disable=W0106:expression-not-assigned, R1702:too-many-branches


from itertools import combinations
import itertools
from multiprocessing import Pool

import numpy as np
import pandas as pd

from .computeScoreMatParallel import computeScoreMatParallel


def computeScoreMat(
        X,
        score_name,
        num_parents,
        verbose,
        num_cores,
        sel_mat,
        pars_score,
        interv_mat,
        interv_data):
    """_summary_

    Args:
        X (_type_): _description_
        score_name (_type_): _description_
        num_parents (_type_): _description_
        output (_type_): _description_
        num_cores (_type_): _description_
        sel_mat (_type_): _description_
        pars_score (_type_): _description_
        interv_mat (_type_): _description_
        interv_data (_type_): _description_

    Returns:
        _type_: _description_
    """

    p = X.shape[1]
    n = X.shape[0]
    row_parents = np.array(
        list(combinations(range(p), num_parents)), dtype=int)

    # XXX
    if verbose:
        print(f". p: {p}")
        print(f". n: {n}")
        print(f". row_parents: {row_parents.flatten()}")

    tt = pd.DataFrame(list(itertools.product(
        range(0, row_parents.shape[0]), range(0, p))),
        columns=['i', 'j'])
    all_node2 = tt['i'].values
    all_i = tt['j'].values

    if num_cores == 1:
        score_mat = np.array(
            [computeScoreMatParallel(
                row_parents, score_name, X, sel_mat, verbose, node2,
                i, pars_score, interv_mat, interv_data)
             for node2, i in zip(all_node2, all_i)])
    else:
        with Pool(num_cores) as pool:
            score_mat = np.array(pool.starmap(computeScoreMatParallel, [(
                row_parents, score_name, X, sel_mat, verbose, node2, i,
                pars_score, interv_mat, interv_data)
                for node2, i in zip(all_node2, all_i)]))

    score_mat = score_mat.reshape(len(row_parents), p)

    init_score = np.empty(p)
    for i in range(p):
        if interv_data:
            X2 = X[~interv_mat[:, i], :]
        else:
            X2 = X
        vartmp = np.var(X2[:, i])
        init_score[i] = -np.log(vartmp)
        score_mat[:, i] -= init_score[i]

    return {'scoreMat': score_mat,
            'rowParents': row_parents,
            'scoreEmptyNodes': init_score}



import importlib
from types import SimpleNamespace

import numpy as np
import pytest

from causalexplain.estimators.cam import train_gam as train_gam_module


def test_train_gam_fallback_when_model_fit_fails(monkeypatch):
    lam_calls = []

    def fake_build(p, num_splines, lam=False):
        lam_calls.append(lam)
        return f"formula-{p}-{num_splines}-{lam}"

    class FakeGAM:
        fail_once = True

        def __init__(self, formula):
            self.formula = formula
            self.statistics_ = {"n_samples": 2, "edof": 1, "p_values": np.array([0.2, 0.4])}
            self._fit_calls = 0

        def fit(self, X, y):
            self._fit_calls += 1
            if FakeGAM.fail_once:
                FakeGAM.fail_once = False
                raise RuntimeError("force fallback")

        def predict(self, X):
            return np.ones((X.shape[0], 1))

    monkeypatch.setattr(train_gam_module, "_build_gam_formula", fake_build)
    monkeypatch.setattr(train_gam_module, "GAM", FakeGAM)

    X = np.array([[0.0], [1.0]])
    y = np.array([0.0, 1.0])
    result = train_gam_module.train_gam(X, y, pars={"num_basis_fcts": 2})

    assert lam_calls == [False, True]
    assert result["residuals"].shape == (2, 1)
    assert np.allclose(result["residuals"].flatten(), y - 1)


def test_build_gam_formula_stacks_terms():
    reloadable = importlib.reload(train_gam_module)
    formula = reloadable._build_gam_formula(3, 5, lam=True)

    assert hasattr(formula, "__add__") or isinstance(formula, SimpleNamespace)



import math

import networkx as nx
import numpy as np
import pandas as pd
from sklearn.preprocessing import StandardScaler

from typing import Optional

# from causalexplain.common.utils import graph_from_dot_file, load_experiment
from ...common import utils

# pylint: disable=E1101:no-member, W0201:attribute-defined-outside-init, W0511:fixme
# pylint: disable=C0103:invalid_name, C0116:missing-function-docstring
# pylint: disable=R0913:too-many-arguments, R1702:too-many-branches
# pylint: disable=R0914:too-many-locals, R0915:too-many-statements


class Knowledge:
    """
    This class collects everything we know about each edge in the proposed graph
    in terms of the following properties:

    - origin: the origin node
    - target: the target node
    - is_edge: whether the edge is in the reference graph
    - is_root_cause: whether the origin is a root cause
    - is_leaf_node: whether the origin is a leaf node
    - correlation: the correlation between the individual SHAP values and the origin node
    - KS_pval: the p-value of the Kolmogorov-Smirnov test between the origin and the target
    - shap_edge: whether the edge is in the graph constructed after evaluating mean
        SHAP values.
    - shap_skedastic_pval: the p-value of the skedastic test for the SHAP values
    - parent_skedastic_pval: the p-value of the skedastic test for the parent values
    - mean_shap: the mean of the SHAP values between the origin and the target
    - slope_shap: the slope of the linear regression for target vs. SHAP values
    - slope_target: the slope of the linear regression for the target vs. origin values
    - potential_root: whether the origin is a potential root cause
    - regression_err: the regression error of the origin to the target
    - err_contrib: the error contribution of the origin to the target
    - con_ind_pval: the p-value of the conditional independence test between the origin
        and the target
    """

    def __init__(self, rex: object, ref_graph: nx.DiGraph):
        """
        Arguments:
        ----------
            shaps (ShapEstimator): The shap estimator.
            ref_graph (nx.DiGraph): The reference graph, or ground truth.
        """
        assert rex is not None, "Rex is None"
        assert rex.hierarchies is not None, "Hierarchies is None"
        assert rex.shaps is not None, "ShapEstimator is None"
        assert rex.pi is not None, "PIEstimator is None"

        self.K = 180.0 / math.pi
        self.shaps = rex.shaps
        self.pi = rex.pi
        self.hierarchies = rex.hierarchies
        self.indep = rex.indep
        self.feature_names = rex.feature_names
        self.scoring = rex.models.scoring
        self.ref_graph = ref_graph
        self.G_shap = rex.G_shap
        self.root_causes = rex.root_causes

        self.correlation_th = rex.correlation_th
        if self.correlation_th is not None:
            self.correlated_features = self.hierarchies.correlated_features

    def info(self):
        """Returns a dataframe with the knowledge about each edge in the graph"""
        rows = []
        ci = self.indep.compute_cond_indep_pvals()
        for target in self.feature_names:
            for parent in self.feature_names:
                if target == parent:
                    continue

                if self.correlation_th is not None:
                    if parent in self.correlated_features[target]:
                        continue

                if self.correlation_th is not None:
                    all_features = [f for f in self.feature_names if (
                        f != target) and (f not in self.correlated_features[target])]
                else:
                    all_features = [
                        f for f in self.feature_names if f != target]
                feature_pos = all_features.index(parent)

                sd = self.shaps.shap_discrepancies[target][parent]
                pi = self.pi.pi[target]['importances_mean'][feature_pos]

                b0_s, beta1_s = sd.shap_model.params[0], sd.shap_model.params[1]
                b0_y, beta1_y = sd.parent_model.params[0], sd.parent_model.params[1]
                shap_slope = math.atan(beta1_s)*self.K
                parent_slope = math.atan(beta1_y)*self.K
                rows.append({
                    'origin': parent,
                    'target': target,
                    'is_edge': int((parent, target) in self.ref_graph.edges()),
                    'o_is_root': int(nx.ancestors(self.ref_graph, parent) == set()),
                    't_is_leaf': int(nx.descendants(self.ref_graph, target) == set()),
                    'correlation': self.hierarchies.correlations[target][parent],
                    'shap_corr': sd.shap_correlation,
                    'shap_gof': sd.shap_gof,
                    'ks_pval': sd.ks_pvalue,
                    'shap_edge': int(parent in set(self.G_shap.predecessors(target))),
                    'shap_sk_pval': sd.shap_p_value,
                    'parent_sk_pval': sd.parent_p_value,
                    'mean_shap': self.shaps.shap_mean_values[target][feature_pos],
                    'mean_pi': pi,
                    'slope_shap': shap_slope,
                    'slope_target': parent_slope,
                    'pot_root': int(parent in self.root_causes),
                    'regr_err': self.scoring[feature_pos],
                    'err_contrib': self.shaps.error_contribution.loc[target, parent],
                    'cond_ind_pval': ci[(target, parent)]
                })
        self.results = pd.DataFrame.from_dict(rows)
        return self.results

    def retrieve(self, origin: str, target: str, what: Optional[str] = None):
        """Returns the knowledge about a specific edge"""
        if what is None:
            return self.results[(self.results.origin == origin) &
                                (self.results.target == target)]

        return self.results[(self.results.origin == origin) &
                            (self.results.target == target)][what].values[0]


if __name__ == "__main__":

    # Display Options
    np.set_printoptions(precision=4, linewidth=100)
    pd.set_option('display.precision', 4)
    pd.set_option('display.float_format', '{:.4f}'.format)

    # Paths
    path = "/Users/renero/phd/data/RC3/"
    output_path = "/Users/renero/phd/output/RC3/"
    # experiment_name = 'rex_generated_linear_1'
    experiment_name = 'custom_rex'

    # Read the data
    reference_graph = utils.graph_from_dot_file(f"{path}{experiment_name}.dot")
    data = pd.read_csv(f"{path}{experiment_name}.csv")
    scaler = StandardScaler()
    data = pd.DataFrame(scaler.fit_transform(data), columns=data.columns)

    custom = utils.load_experiment(f"{experiment_name}", output_path)
    custom.is_fitted_ = True
    print(f"Loaded experiment {experiment_name}")

    custom.feature_names = list(data.columns)
    custom.models.score(data)
    custom.knowledge(reference_graph)



import numpy as np
import torch

from causalexplain.estimators.notears import nonlinear


def test_notears_mlp_forward_and_regularizers():
    torch.set_default_dtype(torch.double)
    model = nonlinear.NotearsMLP(dims=[2, 2, 1], bias=True)
    with torch.no_grad():
        model.fc1_pos.weight[:] = 0.5
        model.fc1_neg.weight[:] = 0.1
    x = torch.tensor([[1.0, 2.0], [0.5, -0.5]], dtype=torch.double)
    output = model(x)
    assert output.shape == (2, 2)
    assert model.h_func().dtype == torch.double
    assert model.l2_reg() >= 0
    assert model.fc1_l1_reg() >= 0
    adj = model.fc1_to_adj()
    assert adj.shape == (2, 2)
    assert np.all(adj >= 0)


def test_notears_sobolev_forward_and_adj():
    torch.set_default_dtype(torch.double)
    model = nonlinear.NotearsSobolev(d=2, k=2)
    x = torch.tensor([[0.1, -0.2]], dtype=torch.double)
    out = model(x)
    assert out.shape == (1, 2)
    assert model.h_func().dtype == torch.double
    assert model.l2_reg() >= 0
    assert model.fc1_l1_reg() >= 0
    adj = model.fc1_to_adj()
    assert adj.shape == (2, 2)


def test_dual_ascent_step_uses_model(monkeypatch):
    class DummyOptim:
        def __init__(self, _):
            self.steps = 0

        def step(self, closure):
            self.steps += 1
            closure()

        def zero_grad(self):
            pass

    class DummyModel(torch.nn.Module):
        def __init__(self):
            super().__init__()
            self.w = torch.nn.Parameter(torch.tensor([[0.0]], dtype=torch.double))
            self.h_val = torch.tensor(0.0, dtype=torch.double, requires_grad=True)

        def forward(self, x):
            return x * self.w

        def h_func(self):
            return self.h_val

        def l2_reg(self):
            return torch.sum(self.w ** 2)

        def fc1_l1_reg(self):
            return torch.sum(torch.abs(self.w))

    monkeypatch.setattr(nonlinear, "LBFGSBScipy", DummyOptim)

    model = DummyModel()
    X = np.zeros((2, 1))
    rho, alpha, h_new = nonlinear.dual_ascent_step(
        model, X, lambda1=0.0, lambda2=0.0, rho=1.0, alpha=0.0, h=np.inf, rho_max=10.0)

    assert rho == 1.0
    assert alpha == 0.0
    assert h_new == 0.0


def test_notears_nonlinear_stops_when_constraint_small(monkeypatch):
    calls = {"iterations": 0}

    def fake_dual_ascent(model, X, lambda1, lambda2, rho, alpha, h, rho_max):
        calls["iterations"] += 1
        return rho, alpha, 0.0  # immediately satisfy h <= h_tol

    class DummyModel(torch.nn.Module):
        def forward(self, x):
            return x

        def fc1_to_adj(self):
            return np.array([[0.05, 0.0], [0.0, 0.0]])

    monkeypatch.setattr(nonlinear, "dual_ascent_step", fake_dual_ascent)

    model = DummyModel()
    X = np.zeros((1, 2))
    W_est = nonlinear.notears_nonlinear(
        model, X, lambda1=0.0, lambda2=0.0, max_iter=5, h_tol=1e-8, w_threshold=0.1)

    assert calls["iterations"] == 1
    assert np.array_equal(W_est, np.zeros((2, 2)))



import pandas as pd
import torch

from causalexplain.models._columnar import ColumnsDataset


def test_columns_dataset_shapes_and_length():
    df = pd.DataFrame(
        {
            "target": [0.1, 0.2, 0.3],
            "f1": [1.0, 2.0, 3.0],
            "f2": [4.0, 5.0, 6.0],
        }
    )

    dataset = ColumnsDataset("target", df)

    assert len(dataset) == 3
    assert dataset.features.dtype == torch.float32
    assert dataset.target.dtype == torch.float32

    features, target = dataset[1]
    assert torch.equal(target, torch.tensor([0.2], dtype=torch.float32))
    assert torch.allclose(features, torch.tensor([2.0, 5.0]))



import pandas as pd
import torch
from torch.utils.data import Dataset


class ColumnsDataset(Dataset):
    def __init__(self, target_name, df: pd.DataFrame):
        target = df.loc[:, target_name].values.reshape(-1, 1)
        features = df.drop(target_name, axis=1).values
        self.features = torch.tensor(features, dtype=torch.float32)
        self.target = torch.tensor(target, dtype=torch.float32)

    def __len__(self):
        return len(self.target)

    def __getitem__(self, idx):
        # Align returned feature dtype with the current global default to avoid
        # downstream dtype mismatches (some tests change the default to double).
        features = self.features[idx].to(torch.get_default_dtype())
        return [features, self.target[idx]]



"""
This Python version attempts to replicate the functionality of the R function. 
Here are some key points:

1. We use pandas for data manipulation and sklearn for machine learning components.
2. The bbs function in R is approximated using SplineTransformer from scikit-learn.
3. Instead of mboost_fit, we use GradientBoostingRegressor from scikit-learn.
4. The function returns a dictionary with the same keys as the R version.

Note that this is an approximation, as the exact behavior of bbs and mboost_fit in R 
might differ from the Python implementations. You may need to fine-tune parameters 
or use different libraries for a more exact replication of the R function's behavior.
"""
import numpy as np
import pandas as pd
from sklearn.ensemble import GradientBoostingRegressor
from sklearn.preprocessing import SplineTransformer


def train_GAMboost(X, y, pars=None):
    """_summary_

    Args:
        X (_type_): _description_
        y (_type_): _description_
        pars (_type_, optional): _description_. Defaults to None.

    Returns:
        _type_: _description_
    """
    # Convert X to DataFrame if it's not already
    dat = pd.DataFrame(X)

    # Create spline features for each column
    spline_features = []
    for column in dat.columns:
        spline = SplineTransformer(n_knots=5, degree=3)
        spline_features.append(spline.fit_transform(dat[[column]]))

    # Concatenate all spline features
    X_spline = np.hstack(spline_features)

    # Fit Gradient Boosting model
    gb = GradientBoostingRegressor()
    gb.fit(X_spline, y)

    # Prepare results
    result = {
        'Yfit': gb.predict(X_spline),
        'residuals': y - gb.predict(X_spline),
        'model': gb
    }

    return result



import torch
import numpy as np
import scipy.linalg as slin


class TraceExpm(torch.autograd.Function):
    @staticmethod
    def forward(ctx, input):
        # detach so we can cast to NumPy
        E = slin.expm(input.detach().numpy())
        f = np.trace(E)
        E = torch.from_numpy(E)
        ctx.save_for_backward(E)
        return torch.as_tensor(f, dtype=input.dtype)

    @staticmethod
    def backward(ctx, grad_output):
        E, = ctx.saved_tensors
        grad_input = grad_output * E.t()
        return grad_input


trace_expm = TraceExpm.apply


def main():
    input = torch.randn(20, 20, dtype=torch.double, requires_grad=True)
    assert torch.autograd.gradcheck(trace_expm, input)

    input = torch.tensor([[1, 2], [3, 4.]], requires_grad=True)
    tre = trace_expm(input)
    f = 0.5 * tre * tre
    print('f\n', f.item())
    f.backward()
    print('grad\n', input.grad)


if __name__ == '__main__':
    main()



# Copyright 2021 Juan L Gamella

# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:

# 1. Redistributions of source code must retain the above copyright
# notice, this list of conditions and the following disclaimer.

# 2. Redistributions in binary form must reproduce the above copyright
# notice, this list of conditions and the following disclaimer in the
# documentation and/or other materials provided with the distribution.

# 3. Neither the name of the copyright holder nor the names of its
# contributors may be used to endorse or promote products derived from
# this software without specific prior written permission.

# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
# COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
# ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

"""
Module containing the auxiliary functions used in the
implementation of GES, including the PDAG to CPDAG conversion
algorithm described in Chickering's original GES paper from 2002.
"""

import numpy as np
import itertools

# --------------------------------------------------------------------
# Graph functions for PDAGS


def na(y, x, A):
    """Return all neighbors of y which are adjacent to x in A.

    Parameters
    ----------
    y : int
        the node's index
    x : int
        the node's index
    A : np.array
        the adjacency matrix of the graph, where A[i,j] != 0 => i -> j
        and A[i,j] != 0 & A[j,i] != 0 => i - j.

    Returns
    -------
    nodes : set of ints
        the resulting nodes

    """
    # Treat the queried node as adjacent to itself so it can be returned
    # when it is an undirected neighbor of y.
    return neighbors(y, A) & (adj(x, A) | {x})


def neighbors(i, A):
    """The neighbors of i in A, i.e. all nodes connected to i by an
    undirected edge.

    Parameters
    ----------
    i : int
        the node's index
    A : np.array
        the adjacency matrix of the graph, where A[i,j] != 0 => i -> j
        and A[i,j] != 0 & A[j,i] != 0 => i - j.

    Returns
    -------
    nodes : set of ints
        the neighbor nodes

    """
    return set(np.where(np.logical_and(A[i, :] != 0, A[:, i] != 0))[0])


def adj(i, A):
    """The adjacent nodes of i in A, i.e. all nodes connected by a
    directed or undirected edge.
    Parameters
    ----------
    i : int
        the node's index
    A : np.array
        the adjacency matrix of the graph, where A[i,j] != 0 => i -> j
        and A[i,j] != 0 & A[j,i] != 0 => i - j.

    Returns
    -------
    nodes : set of ints
        the adjacent nodes

    """
    return set(np.where(np.logical_or(A[i, :] != 0, A[:, i] != 0))[0])


def pa(i, A, directed_only=False):
    """The parents of i in A.

    Parameters
    ----------
    i : int
        the node's index
    A : np.array
        the adjacency matrix of the graph, where A[i,j] != 0 => i -> j
        and A[i,j] != 0 & A[j,i] != 0 => i - j.

    directed_only : bool, optional
        if True, only return strictly directed parents (j -> i and
        i -> j does not hold). If False, return all adjacent nodes
        (incoming or outgoing) to match a looser notion of parents
        sometimes used in PDAG utilities.

    Returns
    -------
    nodes : set of ints
        the parent nodes

    """
    if directed_only:
        return set(np.where(np.logical_and(A[:, i] != 0, A[i, :] == 0))[0])
    incoming = set(np.where(A[:, i] != 0)[0])
    outgoing = set(np.where(A[i, :] != 0)[0])
    return (incoming | outgoing) - {i}


def ch(i, A, directed_only=False):
    """The children of i in A.

    Parameters
    ----------
    A : np.array
        the adjacency matrix of the graph, where A[i,j] != 0 => i -> j
        and A[i,j] != 0 & A[j,i] != 0 => i - j.

    directed_only : bool, optional
        if True, only return strictly directed children (i -> j and
        j -> i does not hold). If False, return all adjacent nodes
        (incoming or outgoing) to match a looser notion of children
        sometimes used in PDAG utilities.

    Returns
    -------
    nodes : set of ints
        the children nodes

    """
    if directed_only:
        return set(np.where(np.logical_and(A[i, :] != 0, A[:, i] == 0))[0])
    outgoing = set(np.where(A[i, :] != 0)[0])
    incoming = set(np.where(A[:, i] != 0)[0])
    return (outgoing | incoming) - {i}


def is_clique(S, A):
    """Check if the subgraph of A induced by nodes S is a clique.

    Parameters
    ----------
    S : set of ints
        set containing the nodes' indices
    A : np.array
        the adjacency matrix of the graph, where A[i,j] != 0 => i -> j
        and A[i,j] != 0 & A[j,i] != 0 => i - j.

    Returns
    -------
    is_clique : bool
        if the subgraph induced by S is a clique in A

    """
    S = list(S)
    n = len(S)
    if n <= 1:
        return True
    subgraph = A[S, :][:, S]
    if n == 2:
        # A single edge (in any direction) is enough for a 2-node clique
        return bool(np.any(subgraph != 0))
    undirected = only_undirected(subgraph)
    no_edges = np.sum(undirected != 0)
    return bool(no_edges == n * (n - 1))


def is_dag(A):
    """Checks wether the given adjacency matrix corresponds to a DAG.

    Parameters
    ----------
    A : np.array
        the adjacency matrix of the graph, where A[i,j] != 0 => i -> j.

    Returns
    -------
    is_dag : bool
        if the adjacency corresponds to a DAG
    """
    try:
        topological_ordering(A)
        return True
    except ValueError:
        return False


def topological_ordering(A):
    """Return a topological ordering for the DAG with adjacency matrix A,
    using Kahn's 1962 algorithm.

    Raises a ValueError exception if the given adjacency does not
    correspond to a DAG.

    Parameters
    ----------
    A : np.array
        the adjacency matrix of the graph, where A[i,j] != 0 => i -> j.

    Returns
    -------
    ordering : list of ints
        a topological ordering for the DAG

    Raises
    ------
    ValueError :
        If the given adjacency does not correspond to a DAG.

    """
    # Check that there are no undirected edges
    if only_undirected(A).sum() > 0:
        raise ValueError("The given graph is not a DAG")
    # Run the algorithm from the 1962 paper "Topological sorting of
    # large networks" by AB Kahn
    A = A.copy()
    sinks = list(np.where(A.sum(axis=0) == 0)[0])
    ordering = []
    while len(sinks) > 0:
        i = sinks.pop()
        ordering.append(i)
        for j in ch(i, A, directed_only=True):
            A[i, j] = 0
            if len(pa(j, A, directed_only=True)) == 0:
                sinks.append(j)
    # If A still contains edges there is at least one cycle
    if A.sum() > 0:
        raise ValueError("The given graph is not a DAG")
    else:
        return ordering


def semi_directed_paths(fro, to, A):
    """Return all paths from i to j in A. Note: a path is a sequence
    (a_1,...,a_n) of non-repeating nodes where either a_i -> a_i+1 or
    a_i - a_i+1 are edges in the PDAG A.

    Parameters
    ----------
    fro : int
        the index of the starting node
    to : int
        the index of the target node
    A : np.array
        the adjacency matrix of the graph, where A[i,j] != 0 => i -> j.

    Returns
    -------
    paths : list of lists
        all the paths between the two nodes

    """
    # NOTE: Implemented non-recursively to avoid issues with Python's
    # stack size limit
    stack = [(fro, [], list(ch(fro, A, directed_only=True) |
                            neighbors(fro, A)))]
    paths = []
    # Precompute the nodes that are accessible from each node for a
    # significant increase in speed
    accessible = dict(
        (i, ch(i, A, directed_only=True) | neighbors(i, A))
        for i in range(len(A))
    )
    while len(stack) > 0:
        current_node, visited, to_visit = stack[0]
        if current_node == to:
            paths.append(visited + [current_node])
            stack = stack[1:]
        elif to_visit == []:
            stack = stack[1:]
        else:
            next_node = to_visit.pop()
            next_to_visit = list(
                accessible[next_node] - set(visited) - {current_node})
            stack = [(next_node, visited + [current_node], next_to_visit)] + stack
    return paths


def separates(S, A, B, G):
    """Returns true if the set S separates A from B in G, i.e. if all
    paths in G from nodes in A to nodes in B contain a node in
    S. Exception is raised if S,A and B are not pairwise disjoint.

    Parameters
    ----------
    S : set of ints
        a set of nodes in G
    A : set of ints
        a set of nodes in G
    B : set of ints
        a set of nodes in G
    G : np.array
        the adjacency matrix of the graph, where G[i,j] != 0 => i -> j
        and G[i,j] != 0 and G[j,i] != 0 => i - j.

    Returns
    -------
    separated : bool
        if S separates A from B in G

    """
    # Check that sets are pairwise disjoint
    if len(A & B) or len(A & S) or len(B & S):
        raise ValueError(
            "The sets S=%s,A=%s and B=%s are not pairwise disjoint" % (S, A, B))
    for a in A:
        for b in B:
            for path in semi_directed_paths(a, b, G):
                if set(path) & S == set():
                    return False
    return True


def chain_component(i, G):
    """Return all nodes in the connected component of node i after
    dropping all directed edges in G.

    Parameters
    ----------
    i : int
        the node's index
    G : np.array
        the adjacency matrix of the graph, where G[i,j] != 0 => i -> j
        and G[i,j] != 0 and G[j,i] != 0 => i - j

    Returns
    -------
    visited : set of ints
        the nodes in the chain component of i

    """
    A = only_undirected(G)
    visited = set()
    to_visit = {i}
    # NOTE: Using a breadth-first search
    while len(to_visit) > 0:
        for j in to_visit:
            visited.add(j)
            to_visit = (to_visit | neighbors(j, A)) - visited
    return visited


def induced_subgraph(S, G):
    """Remove all edges which are not between nodes in S.

    Parameters
    ----------
    S : set of ints
        a set of node indices
    G : np.array
        the adjacency matrix of the graph, where G[i,j] != 0 => i -> j
        and G[i,j] != 0 and G[j,i] != 0 => i - j

    Returns
    -------
    subgraph : np.array
       the adjacency matrix of the resulting graph where all edges
       between nodes not in S are removed. Note that this is not
       really the subgraph, as the nodes not in S still appear as
       disconnected nodes.
    """
    mask = np.zeros_like(G, dtype=bool)
    mask[list(S), :] = True
    mask = np.logical_and(mask, mask.T)
    subgraph = np.zeros_like(G)
    subgraph[mask] = G[mask]
    return subgraph


def vstructures(A):
    """
    Return the v-structures of a DAG or PDAG, given its adjacency matrix.

    Parameters
    ----------
    A : np.array
        The adjacency of the (P)DAG, where A[i,j] != 0 => i->j

    Returns
    -------
    vstructs : set()
        the set of v-structures, where every v-structure is a three
        element tuple, e.g. (i,j,k) represents the v-structure
        i -> j <- k, where i < j for consistency.

    """
    # Construct the graph with only the directed edges
    dir_A = only_directed(A)
    # Search for colliders in the graph with only directed edges
    colliders = np.where((dir_A != 0).sum(axis=0) > 1)[0]
    # For each collider, and all pairs of parents, check if the
    # parents are adjacent in A
    vstructs = []
    for c in colliders:
        for (i, j) in itertools.combinations(pa(c, A, directed_only=True), 2):
            if A[i, j] == 0 and A[j, i] == 0:
                # Ordering might be defensive here, as
                # itertools.combinations already returns ordered
                # tuples; motivation is to not depend on their feature
                vstruct = (i, c, j) if i < j else (j, c, i)
                vstructs.append(vstruct)
    return set(vstructs)


def only_directed(P):
    """
    Return the graph with the same nodes as P and only its directed edges.

    Parameters
    ----------
    P : np.array
        adjacency matrix of a graph

    Returns
    -------
    G : np.array
        adjacency matrix of the graph with the same nodes as P and
        only its directed edges

    """
    mask = np.logical_and(P != 0, P.T == 0)
    G = np.zeros_like(P)
    # set to the same values in case P is a weight matrix and there is
    # interest in maintaining the weights
    G[mask] = P[mask]
    return G


def only_undirected(P):
    """
    Return the graph with the same nodes as P and only its undirected edges.

    Parameters
    ----------
    P : np.array
        adjacency matrix of a graph

    Returns
    -------
    G : np.array
        adjacency matrix of the graph with the same nodes as P and
        only its undirected edges

    """
    mask = np.logical_and(P != 0, P.T != 0)
    G = np.zeros_like(P)
    # set to the same values in case P is a weight matrix and there is
    # interest in maintaining the weights
    G[mask] = P[mask]
    return G


def skeleton(A):
    """Return the skeleton of a given graph.

    Parameters
    ----------
    A : np.array
        adjacency matrix of a graph

    Returns
    -------
    S : np.array
        adjacency matrix of the skeleton, i.e. the graph resulting
        from dropping all edge orientations

    """
    return ((A + A.T) != 0).astype(int)


def is_consistent_extension(G, P, debug=False):
    """Returns True if the DAG G is a consistent extension of the PDAG
    P. Will raise a ValueError exception if the graph G is not a DAG
    (i.e. cycles or undirected edges).

    Parameters
    ----------
    G : np.array
        the adjacency matrix of DAG
    P : np.array
        the adjacency matrix of PDAG
    debug : bool
        if debugging traces should be outputted

    Returns
    -------
    consistent : bool
        True if G is a consistent extension of P (see below)

    """
    # G is a consistent extension of P iff
    #   0. it is a DAG
    #   1. they have the same v-structures,
    #   2. they have the same skeleton, and
    #   3. every directed edge in P has the same direction in G
    if not is_dag(G):
        raise ValueError("G is not a DAG")
    # v-structures
    same_vstructures = vstructures(P) == vstructures(G)
    # skeleton
    same_skeleton = (skeleton(P) == skeleton(G)).all()
    # same orientation
    directed_P = only_directed(P)
    same_orientation = G[directed_P != 0].all()  # No need to check
    # transpose as G is
    # guaranteed to have
    # no undirected edges
    if debug:
        print("v-structures (%s) (P,G): " %
              same_vstructures, vstructures(P), vstructures(G))
        print("skeleton (%s) (P,G): " %
              same_skeleton, skeleton(P), skeleton(G))
        print("orientation (%s) (P,G): " % same_orientation, P, G)
    return same_vstructures and same_orientation and same_skeleton

# --------------------------------------------------------------------
# Functions for PDAG to CPDAG conversion

# The following functions implement the conversion from PDAG to
# CPDAG that is carried after each transition to a different
# equivalence class, after the selection and application of the
# highest scoring insert/delete/turn operator. It consists of the
# succesive application of three algorithms, all described in
# Appendix C (pages 552,553) of Chickering's 2002 GES paper
# (www.jmlr.org/papers/volume3/chickering02b/chickering02b.pdf).
#
# The algorithms are:

#   1. Obtaining a consistent extension of a PDAG, implemented in
#   the function pdag_to_dag.
#
#   2. Obtaining a total ordering of the edges of the extension
#   resulting from (1). It is summarized in Fig. 13 of
#   Chickering's paper and implemented in the function
#   order_edges.
#
#   3. Labelling the edges as compelled or reversible, by which we
#   can easily obtain the CPDAG. It is summarized in Fig. 14 of
#   Chickering's paper and implemented in the function label_edges.

# The above are put together in the function pdag_to_cpdag

# NOTE!!!: Algorithm (1) is from the 1992 paper "A simple
# algorithm to construct a consistent extension of a partially
# oriented graph" by Dorit Dor and Michael Tarsi. There is an
# ERROR in the summarized version in Chickering's paper. In
# particular, the condition that N_x U Pa_x is a clique is not
# equivalent to the condition from Dor & Torsi that every neighbor
# of X should be adjacent to all of X's adjacent nodes. The
# condition summarized in Chickering is more restrictive (i.e. it
# also asks that the parents of X are adjacent to each other), but
# this only results in an error for some graphs, and was only
# uncovered during exhaustive testing.

# The complete pipeline: pdag -> dag -> ordered -> labelled -> cpdag


def pdag_to_cpdag(pdag):
    """
    Transform a PDAG into its corresponding CPDAG. Returns a ValueError
    exception if the given PDAG does not admit a consistent extension.

    Parameters
    ----------
    pdag : np.array
        the adjacency matrix of a given PDAG where pdag[i,j] != 0 if i
        -> j and i - j if also pdag[j,i] != 0.

    Returns
    -------
    cpdag : np.array
        the adjacency matrix of the corresponding CPDAG

    """
    # 1. Obtain a consistent extension of the pdag
    dag = pdag_to_dag(pdag)
    # 2. Recover the cpdag
    return dag_to_cpdag(dag)

# dag -> ordered -> labelled -> cpdag


def dag_to_cpdag(G):
    """
    Return the completed partially directed acyclic graph (CPDAG) that
    represents the Markov equivalence class of a given DAG. Returns a
    ValueError exception if the given graph is not a DAG.

    Parameters
    ----------
    G : np.array
        the adjacency matrix of the given graph, where G[i,j] != 0 iff i -> j

    Returns
    -------
    cpdag : np.array
        the adjacency matrix of the corresponding CPDAG

    """
    # 1. Perform a total ordering of the edges
    ordered = order_edges(G)
    # 2. Label edges as compelled or reversible
    labelled = label_edges(ordered)
    # 3. Construct CPDAG
    cpdag = np.zeros_like(labelled)
    # set compelled edges
    cpdag[labelled == 1] = labelled[labelled == 1]
    # set reversible edges
    fros, tos = np.where(labelled == -1)
    for (x, y) in zip(fros, tos):
        cpdag[x, y], cpdag[y, x] = 1, 1
    return cpdag


def pdag_to_dag(P, debug=False):
    """
    Find a consistent extension of the given PDAG. Return a ValueError
    exception if the PDAG does not admit a consistent extension.

    Parameters
    ----------
    P : np.array
        adjacency matrix representing the PDAG connectivity, where
        P[i,j] = 1 => i->j
    debug : bool, optional
        if debugging traces should be printed

    Returns
    -------
    G : np.array
        the adjacency matrix of a DAG which is a consistent extension
        (i.e. same v-structures and skeleton) of P.

    """
    G = only_directed(P)
    indexes = list(range(len(P)))  # To keep track of the real variable
    # indexes as we remove nodes from P
    while P.size > 0:
        print(P) if debug else None
        print(indexes) if debug else None
        # Select a node which
        #   1. has no outgoing edges in P (i.e. childless, is a sink)
        #   2. all its neighbors are adjacent to all its adjacent nodes
        found = False
        i = 0
        while not found and i < len(P):
            # Check condition 1
            sink = len(ch(i, P, directed_only=True)) == 0
            # Check condition 2
            n_i = neighbors(i, P)
            adj_i = adj(i, P)
            adj_neighbors = np.all([adj_i - {y} <= adj(y, P) for y in n_i])
            print("   i:", i, ": n=", n_i, "adj=", adj_i,
                  "ch=", ch(i, P, directed_only=True)) if debug else None
            found = sink and adj_neighbors
            # If found, orient all incident undirected edges and
            # remove i from the subgraph
            if found:
                print("  Found candidate %d (%d)" %
                      (i, indexes[i])) if debug else None
                # Orient all incident undirected edges
                real_i = indexes[i]
                real_neighbors = [indexes[j] for j in n_i]
                for j in real_neighbors:
                    G[j, real_i] = 1
                # Remove i and its incident (directed and undirected edges)
                all_but_i = list(set(range(len(P))) - {i})
                P = P[all_but_i, :][:, all_but_i]
                indexes.remove(real_i)  # to keep track of the real
                # variable indices
            else:
                i += 1
        # A node which satisfies conditions 1,2 exists iff the
        # PDAG admits a consistent extension
        if not found:
            raise ValueError("PDAG does not admit consistent extension")
    return G


def order_edges(G):
    """
    Find a total ordering of the edges in DAG G, as an intermediate
    step to obtaining the CPDAG representing the Markov equivalence class to
    which it belongs. Raises a ValueError exception if G is not a DAG.

    Parameters
    ----------
    G : np.array
        the adjacency matrix of a graph G, where G[i,j] != 0 iff i -> j.

    Returns
    -------
    ordered : np.array
       the adjacency matrix of the graph G, but with labelled edges,
       i.e. i -> j is has label x iff ordered[i,j] = x.

    """
    if not is_dag(G):
        raise ValueError("The given graph is not a DAG")
    # i.e. if i -> j, then i appears before j in order
    order = topological_ordering(G)
    # You can check the above by seeing that np.all([i == order[pos[i]] for i in range(p)]) is True
    # Unlabelled edges as marked with -1
    ordered = (G != 0).astype(int) * -1
    i = 1
    while (ordered == -1).any():
        # let y be the lowest ordered node that has an unlabelled edge
        # incident to it
        froms, tos = np.where(ordered == -1)
        with_unlabelled = np.unique(np.hstack((froms, tos)))
        y = sort(with_unlabelled, reversed(order))[0]
        # let x be the highest ordered node s.t. the edge x -> y
        # exists and is unlabelled
        unlabelled_parents_y = np.where(ordered[:, y] == -1)[0]
        x = sort(unlabelled_parents_y, order)[0]
        ordered[x, y] = i
        i += 1
    return ordered


def label_edges(ordered):
    """Given a DAG with edges labelled according to a total ordering,
    label each edge as being compelled or reverisble.

    Parameters
    ----------
    ordered : np.array
        the adjacency matrix of a graph, with the edges labelled
        according to a total ordering.

    Returns
    -------
    labelled : np.array
        the adjacency matrix of G but with labelled edges, where
          - labelled[i,j] = 1 iff i -> j is compelled, and
          - labelled[i,j] = -1 iff i -> j is reversible.

    """
    # Validate the input
    if not is_dag(ordered):
        raise ValueError("The given graph is not a DAG")
    no_edges = (ordered != 0).sum()
    if sorted(ordered[ordered != 0]) != list(range(1, no_edges + 1)):
        raise ValueError("The ordering of edges is not valid:",
                         ordered[ordered != 0])
    # define labels: 1: compelled, -1: reversible, -2: unknown
    COM, REV, UNK = 1, -1, -2
    labelled = (ordered != 0).astype(int) * UNK
    # while there are unknown edges
    while (labelled == UNK).any():
        # print(labelled)
        # let (x,y) be the unknown edge with lowest order
        # (i.e. appears last in the ordering, NOT has smalles label)
        # in ordered
        unknown_edges = (ordered * (labelled == UNK).astype(int)).astype(float)
        unknown_edges[unknown_edges == 0] = -np.inf
        # print(unknown_edges)
        (x, y) = np.unravel_index(np.argmax(unknown_edges), unknown_edges.shape)
        # print(x,y)
        # iterate over all edges w -> x which are compelled
        Ws = np.where(labelled[:, x] == COM)[0]
        end = False
        for w in Ws:
            # if w is not a parent of y, label all edges into y as
            # compelled, and finish this pass
            if labelled[w, y] == 0:
                labelled[list(pa(y, labelled, directed_only=True)), y] = COM
                end = True
                break
            # otherwise, label w -> y as compelled
            else:
                labelled[w, y] = COM
        if not end:
            # if there exists an edge z -> y such that z != x and z is
            # not a parent of x, label all unknown edges (this
            # includes x -> y) into y with compelled; label with
            # reversible otherwise.
            z_exists = len(
                pa(y, labelled, directed_only=True)
                - {x}
                - pa(x, labelled, directed_only=True)
            ) > 0
            unknown = np.where(labelled[:, y] == UNK)[0]
            assert x in unknown
            labelled[unknown, y] = COM if z_exists else REV
    return labelled

# --------------------------------------------------------------------
# General utilities

# Very fast way to generate a cartesian product of input arrays
# Credit: https://gist.github.com/hernamesbarbara/68d073f551565de02ac5


def cartesian(arrays, out=None, dtype=np.byte):
    """
    Generate a cartesian product of input arrays.

    Parameters
    ----------
    arrays : list of array-like
        1-D arrays to form the cartesian product of.
    out : ndarray
        Array to place the cartesian product in.

    Returns
    -------
    out : ndarray
        2-D array of shape (M, len(arrays)) containing cartesian products
        formed of input arrays.

    Examples
    --------
    >>> cartesian(([1, 2, 3], [4, 5], [6, 7]))
    array([[1, 4, 6],
           [1, 4, 7],
           [1, 5, 6],
           [1, 5, 7],
           [2, 4, 6],
           [2, 4, 7],
           [2, 5, 6],
           [2, 5, 7],
           [3, 4, 6],
           [3, 4, 7],
           [3, 5, 6],
           [3, 5, 7]], dtype=int8)

    """
    arrays = [np.asarray(x) for x in arrays]

    n = np.prod([x.size for x in arrays])
    if out is None:
        out = np.zeros([n, len(arrays)], dtype=dtype)

    m = int(n / arrays[0].size)
    out[:, 0] = np.repeat(arrays[0], m)
    if arrays[1:]:
        cartesian(arrays[1:], out=out[0:m, 1:])
        for j in range(1, arrays[0].size):
            out[j * m:(j + 1) * m, 1:] = out[0:m, 1:]
    return out


def sort(L, order=None):
    """Sort the elements in an iterable according to its pre-defined
    'sorted' function, or according to a given order: i will precede j
    if i precedes j in the order.

    Parameters
    ----------
    L : iterable
        the iterable to be sorted
    order : iterable or None, optional
        a given ordering. In the sorted result, i will precede j if i
        precedes j in order. If None, the predefined 'sorted' function
        of the iterator will be used. Defaults to None.

    Returns
    -------
    ordered : list
        a list containing the elements of L, sorted from lesser to
        greater or according to the given order.

    """
    L = list(L)
    if order is None:
        return sorted(L)
    else:
        order = list(order)
        pos = np.zeros(len(order), dtype=int)
        pos[order] = range(len(order))
        positions = [pos[l] for l in L]
        return [tup[1] for tup in sorted(zip(positions, L))]


def subsets(S):
    """
    Return an iterator with all possible subsets of the set S.

    Parameters
    ----------
    S : set
        a given set

    Returns
    -------
    subsets : iterable
        an iterable over the subsets of S, in increasing size

    """
    subsets = []
    for r in range(len(S) + 1):
        subsets += [set(ss) for ss in itertools.combinations(S, r)]
    return subsets


def member(L, A):
    """
    Return the index of the first appearance of array A in L.

    Parameters
    ----------
    L : list of np.array
        list on which to perform the search
    A : np.array
        the target array

    Returns
    -------
    position : int or None
        the index of the first appearance of array A in list L, or
        None if A is not in L.

    """
    for i, B in enumerate(L):
        if (A == B).all():
            return i
    return None


def delete(array, mask, axis=None):
    """Wrapper for numpy.delete, which adapts the call depending on the
    numpy version (the API changed on 1.19.0)

    Parameters
    ----------
    array : array_like
        Input array.
    mask : boolean array
        Specifies the sub-arrays to remove along the given axis
    axis : int
        The axis along which to delete the subarrays specified by mask

    Returns
    -------
    out : ndarray
        a copy of array with the elements specified by mask removed

    """
    if np.version.version < '1.19.0':
        idx = np.where(mask)[0]
        return np.delete(array, idx, axis)
    else:
        return np.delete(array, mask, axis)



import numpy as np
import pytest
import pandas as pd

from causalexplain.independence.edge_orientation import get_edge_orientation


def test_get_edge_orientation():
    # Create a mock DataFrame
    data = pd.DataFrame({
        'x': np.random.rand(100),
        'y': np.random.rand(100)
    })

    # Test with default parameters
    result = get_edge_orientation(data, 'x', 'y')
    assert isinstance(result, int)
    assert result in [-1, 0, 1]

    # Test with custom parameters
    result = get_edge_orientation(data, 'x', 'y', iters=50, method='gam', verbose=True)
    assert isinstance(result, int)
    assert result in [-1, 0, 1]

    # Test with invalid method
    with pytest.raises(ValueError):
        get_edge_orientation(data, 'x', 'y', method='invalid')

    # Test with non-existent columns
    with pytest.raises(KeyError):
        get_edge_orientation(data, 'a', 'b')



import numpy as np

from causalexplain.estimators.cam import train_linear


def test_train_linear_predictions_and_residuals():
    X = np.array([[1.0, 2.0], [2.0, 0.0], [3.0, 4.0]])
    y = np.array([3.0, 1.0, 7.0])

    result = train_linear.train_linear(X, y)

    assert set(result) == {"Yfit", "residuals", "model"}
    assert result["Yfit"].shape == (3, 1)
    assert result["residuals"].shape == (3, 1)
    assert np.allclose(result["residuals"].flatten() + result["Yfit"].flatten(), y)



import numpy as np
import pandas as pd
import pytest

from causalexplain.independence import regressors


class DummyGPR:
    def __init__(self, *args, **kwargs):
        self.fitted = False

    def fit(self, x, y):
        self.fitted = True

    def predict(self, x):
        return np.zeros((x.shape[0],))


class DummyGAM:
    def gridsearch(self, x, y, progress=False):
        return self

    def predict(self, x):
        return np.ones_like(x).flatten()


def test_fit_and_get_residuals_branches(monkeypatch):
    monkeypatch.setattr(regressors, "gpr", lambda normalize_y: DummyGPR())
    monkeypatch.setattr(regressors, "LinearGAM", DummyGAM)
    monkeypatch.setattr(np.random, "normal", lambda loc, scale, size: np.zeros(size))

    X = np.array([0.0, 1.0, 2.0])
    Y = np.array([1.0, 2.0, 3.0])

    res_gpr = regressors.fit_and_get_residuals(X, Y, method="gpr")
    assert res_gpr.shape == (3, 1)
    assert np.allclose(res_gpr, Y.reshape(-1, 1))

    res_gam = regressors.fit_and_get_residuals(X, Y, method="gam")
    assert res_gam.shape == (3, 1)
    assert np.allclose(res_gam, X.reshape(-1, 1) - 1.0)

    with pytest.raises(ValueError):
        regressors.fit_and_get_residuals(X, Y, method="bad")


def test_run_feature_selection_uses_hsic(monkeypatch):
    class DummyHSIC:
        def fit(self, X, y):
            self.stat = np.array([[0.0, 0.5], [0.5, 1.0]])
            return self

    monkeypatch.setattr(regressors, "HSIC", DummyHSIC)
    df = pd.DataFrame({"a": [0.0, 1.0], "b": [1.0, 0.0], "target": [1, 0]})
    stat = regressors.run_feature_selection(df, "target")
    assert np.allclose(stat, [[0.0, 0.5], [0.5, 1.0]])



import networkx as nx

from causalexplain.estimators.pc.pdag import PDAG


def test_copy_and_to_dag_orientation():
    pdag = PDAG(directed_ebunch=[("X", "Y")], undirected_ebunch=[("Y", "Z")])

    copied = pdag.copy()
    assert copied.directed_edges == pdag.directed_edges
    assert copied.undirected_edges == pdag.undirected_edges

    dag = pdag.to_dag()
    assert isinstance(dag, nx.DiGraph)
    assert dag.has_edge("X", "Y")
    assert (dag.has_edge("Y", "Z") or dag.has_edge("Z", "Y"))



"""
Explainability techniques used for causal discovery.

This module contains various techniques and tools for explaining and 
interpreting causal discovery results:

- shapley: Implements Shapley value-based methods for attributing importance to
  features in causal models.
- regression_quality: Provides metrics and tools for assessing the quality of
  regression models used in causal discovery.
- perm_importance: Implements permutation importance methods for feature
  importance in causal models.
- hierarchies: Contains tools for analyzing and visualizing hierarchical
  structures in causal relationships.

These submodules offer a range of approaches to enhance the interpretability
and understanding of causal discovery results, aiding in the validation and
refinement of causal models.
"""

from . import shapley
from . import regression_quality
from . import perm_importance
from . import hierarchies

__all__ = [
    "shapley",
    "regression_quality",
    "perm_importance",
    "hierarchies",
]



"""
Methods in this module perform HSIC independence test, and compute the HSIC
value and statistic.

After checking that most implementations fail to provide consistent results,
I decided to take the implementation from the DoWhy package.
"""
from dataclasses import dataclass
from dowhy.gcm.independence_test import independence_test

@dataclass
class HSIC_Values:
    hsic: float = 0.
    p_value: float = 0.
    stat: float = 0.
    independence: bool = False

    def __init__(self, hsic, p_value, stat, independence):
        self.hsic = hsic
        self.p_value = p_value
        self.stat = stat
        self.independence = independence

    def __str__(self):
        s = f"HSIC...: {self.hsic:.6g}\n"
        s += f"p_value: {self.p_value:.6g}\n"
        s += f"stat...: {self.stat:.6g}\n"
        s += f"indep..: {self.independence}"
        return s


class HSIC:
    """
    Provides a class sklearn-type interface to the Hsic methods in this DoWhy.
    """
    def __init__(self):
        self.p_val: float = 0.0
        self.h_val: float = 0.0
        self.h_stat: float = 0.0
        self.h_indep: bool = False

    def fit(self, X, Y):
        self.p_val = independence_test(X, Y, conditioned_on=None, method="kernel")
        self.h_indep = self.p_val > 0.05
        return HSIC_Values(self.h_val, self.p_value, self.h_stat, self.h_indep)

    @property
    def p_value(self):
        return self.p_val

    @property
    def hsic(self):
        return self.h_val

    @property
    def stat(self):
        return self.h_stat

    @property
    def independence(self):
        return self.h_indep



"""
Python implementation of the LiNGAM algorithms.
The LiNGAM Project: https://sites.google.com/site/sshimizu06/lingam
"""
import numpy as np
from sklearn.linear_model import LassoLarsIC, LinearRegression


def find_all_paths(dag, from_index, to_index, min_causal_effect=0.0):
    """Find all paths from point to point in DAG.

    Parameters
    ----------
    dag : array-like, shape (n_features, n_features)
        The adjacency matrix to fine all paths, where n_features is the number of features.
    from_index : int
        Index of the variable at the start of the path.
    to_index : int
        Index of the variable at the end of the path.
    min_causal_effect : float, optional (default=0.0)
        Threshold for detecting causal direction.
        Causal directions with absolute values of causal effects less than ``min_causal_effect`` are excluded.

    Returns
    -------
    paths : array-like, shape (n_paths)
        List of found path, where n_paths is the number of paths.
    effects : array-like, shape (n_paths)
        List of causal effect, where n_paths is the number of paths.
    """
    # Extract all edges using the standard orientation dag[i, j] = i -> j
    edges = np.argwhere(np.abs(np.nan_to_num(dag)) > min_causal_effect)

    # Aggregate edges by start point (outgoing adjacency)
    to_indices = []
    for i in range(dag.shape[0]):
        adj_list = edges[edges[:, 0] == i][:, 1].tolist()
        to_indices.append(adj_list)

    # DFS
    paths = []
    stack = [from_index]
    stack_to_indice = [to_indices[from_index]]
    while stack:
        if len(stack) > dag.shape[0]:
            raise ValueError(
                "Unable to find the path because a cyclic graph has been specified."
            )

        cur_index = stack[-1]
        to_indice = stack_to_indice[-1]

        if cur_index == to_index:
            paths.append(stack.copy())
            stack.pop()
            stack_to_indice.pop()
        else:
            if len(to_indice) > 0:
                next_index = to_indice.pop(0)
                stack.append(next_index)
                stack_to_indice.append(to_indices[next_index].copy())
            else:
                stack.pop()
                stack_to_indice.pop()

    # Calculate the causal effect for each path
    effects = []
    for p in paths:
        coefs = [dag[p[i], p[i + 1]] for i in range(len(p) - 1)]
        effects.append(np.cumprod(coefs)[-1])

    return paths, effects


def predict_adaptive_lasso(X, predictors, target, gamma=1.0):
    """Predict with Adaptive Lasso.

    Parameters
    ----------
    X : array-like, shape (n_samples, n_features)
        Training data, where n_samples is the number of samples
        and n_features is the number of features.
    predictors : array-like, shape (n_predictors)
        Indices of predictor variable.
    target : int
        Index of target variable.

    Returns
    -------
    coef : array-like, shape (n_features)
        Coefficients of predictor variable.
    """
    lr = LinearRegression()
    lr.fit(X[:, predictors], X[:, target])
    weight = np.power(np.abs(lr.coef_), gamma)
    reg = LassoLarsIC(criterion="bic")
    reg.fit(X[:, predictors] * weight, X[:, target])
    return reg.coef_ * weight



"""
Metrics module for evaluating causal graphs.

This module provides various metrics and comparison tools for causal graphs,
including:
- SID (Structural Intervention Distance)
- Graph comparison utilities
"""

from . import SID
from . import compare_graphs

__all__ = [
    'SID',
    'compare_graphs',
]



'''
Created on 2013/01/26

@author: myamada
'''
import numpy as np


def rbf_dot(X, deg):
    # Set kernel size to median distance between points, if no kernel specified
    if X.ndim == 1:
        X = X[:, np.newaxis]
    m = X.shape[0]
    G = np.sum(X * X, axis=1)[:, np.newaxis]
    Q = np.tile(G, (1, m))
    H = Q + Q.T - 2.0 * np.dot(X, X.T)
    if deg == -1:
        dists = (H - np.tril(H)).flatten()
        deg = np.sqrt(0.5 * np.median(dists[dists > 0]))
    H = np.exp(-H / 2.0 / (deg ** 2))

    return H


def kernel_Delta_norm(xin1, xin2):
    n1 = xin1.shape[1]
    n2 = xin2.shape[1]

    K = np.zeros((n1, n2))
    ulist = np.unique(xin1)

    for ind in ulist:
        c1 = np.sqrt(np.sum(xin1 == ind))
        c2 = np.sqrt(np.sum(xin2 == ind))
        ind1 = np.where(xin1 == ind)[1]
        ind2 = np.where(xin2 == ind)[1]
        K[np.ix_(ind1, ind2)] = 1 / c1 / c2

    return K


def kernel_Delta(xin1, xin2):
    n1 = xin1.shape[1]
    n2 = xin2.shape[1]

    K = np.zeros((n1, n2))
    ulist = np.unique(xin1)

    for ind in ulist:
        ind1 = np.where(xin1 == ind)[1]
        ind2 = np.where(xin2 == ind)[1]
        K[np.ix_(ind1, ind2)] = 1

    return K


def kernel_Gaussian(xin1, xin2, sigma):
    n1 = xin1.shape[1]
    n2 = xin2.shape[1]

    xin12 = np.sum(np.power(xin1, 2), 0)
    xin22 = np.sum(np.power(xin2, 2), 0)

    dist2 = np.tile(xin22, (n1, 1)) + np.tile(xin12,
                                              (n2, 1)).transpose() - 2 * np.dot(
        xin1.T, xin2)
    K = np.exp(-dist2 / (2 * np.power(sigma, 2)))

    return K



import networkx as nx
import numpy as np
import pandas as pd
import pytest

from causalexplain.estimators.fci import fci as fci_module


def test_filter_causes_only_detects_direct_relationships():
    model = fci_module.FCI(name="filter-test")
    adjacency = np.array([
        [0, -1, 2],
        [1, 0, 0],
        [1, -1, 0],
    ])

    filtered = model.filter_causes_only(adjacency)

    expected = np.zeros_like(adjacency)
    expected[0, 1] = 1  # -1 / 1 combination -> direct edge 0->1
    expected[0, 2] = 1  # 2 / 1 combination -> direct edge 0->2

    assert np.array_equal(filtered, expected)


def test_fit_predict_causes_only_filters_and_sets_attributes(monkeypatch):
    X = pd.DataFrame({"a": [0.1, 0.2], "b": [0.2, 0.3]})
    ref_graph = object()
    fci_called = {}

    class GraphStub:
        def __init__(self, matrix):
            self.graph = matrix

    def fake_fci(values, indep_test_method, alpha, depth, max_path_length,
                 verbose, show_progress, background_knowledge, node_names):
        fci_called["kwargs"] = {
            "indep_test_method": indep_test_method,
            "alpha": alpha,
            "depth": depth,
            "max_path_length": max_path_length,
            "verbose": verbose,
            "show_progress": show_progress,
            "background_knowledge": background_knowledge,
            "node_names": node_names,
        }
        matrix = np.array([[0, -1], [1, 0]])
        return GraphStub(matrix), ["edge"]

    adjacency_inputs = {}

    def fake_graph_from_adjacency(matrix, node_labels):
        adjacency_inputs["matrix"] = matrix
        adjacency_inputs["labels"] = node_labels
        return "graph_obj"

    metrics_called = {}

    def fake_evaluate_graph(ref, dag):
        metrics_called["args"] = (ref, dag)
        return {"score": 0.9}

    monkeypatch.setattr(fci_module, "fci", fake_fci)
    monkeypatch.setattr(fci_module.utils, "graph_from_adjacency",
                        fake_graph_from_adjacency)
    monkeypatch.setattr(fci_module, "evaluate_graph", fake_evaluate_graph)

    model = fci_module.FCI(
        name="experiment",
        independence_test_method="kci",
        alpha=0.01,
        depth=4,
        max_path_length=7,
        background_knowledge="bk",
        node_names=["a", "b"],
        causes_only=True,
    )

    result = model.fit_predict(X=X, ref_graph=ref_graph)

    assert result is model
    assert fci_called["kwargs"] == {
        "indep_test_method": "kci",
        "alpha": 0.01,
        "depth": 4,
        "max_path_length": 7,
        "verbose": False,
        "show_progress": False,
        "background_knowledge": "bk",
        "node_names": ["a", "b"],
    }
    assert np.array_equal(adjacency_inputs["matrix"], np.array([[0, 1], [0, 0]]))
    assert adjacency_inputs["labels"] == ["a", "b"]
    assert metrics_called["args"] == (ref_graph, "graph_obj")
    assert model.metrics == {"score": 0.9}
    assert model.dag == "graph_obj"


def test_fit_predict_without_causes_only_retains_full_adjacency(monkeypatch):
    X = pd.DataFrame({"x": [1, 2], "y": [3, 4]})

    class GraphStub:
        def __init__(self, matrix):
            self.graph = matrix

    adjacency = np.array([[0, 2], [2, 0]])

    def fake_fci(values, **_):
        return GraphStub(adjacency), ["edge"]

    adjacency_inputs = {}

    def fake_graph_from_adjacency(matrix, node_labels):
        adjacency_inputs["matrix"] = matrix
        adjacency_inputs["labels"] = node_labels
        return "full_graph"

    monkeypatch.setattr(fci_module, "fci", fake_fci)
    monkeypatch.setattr(fci_module.utils, "graph_from_adjacency",
                        fake_graph_from_adjacency)
    monkeypatch.setattr(fci_module, "evaluate_graph",
                        lambda ref_graph, dag: {"ref": ref_graph, "dag": dag})

    model = fci_module.FCI(name="no-filter", causes_only=False)
    result = model.fit_predict(X=X, ref_graph="ref")

    assert result.dag == "full_graph"
    assert np.array_equal(adjacency_inputs["matrix"], adjacency)
    assert adjacency_inputs["labels"] == ["x", "y"]
    assert model.metrics == {"ref": "ref", "dag": "full_graph"}


@pytest.mark.parametrize(
    "with_dag,metrics_available",
    [
        (True, True),
        (False, False),
    ],
)
def test_main_prints_edges_and_metrics(monkeypatch, capsys, with_dag, metrics_available):
    dataset_name = "sample"
    dataframe = pd.DataFrame({"x": [1], "y": [2]})

    monkeypatch.setattr(fci_module.pd, "read_csv", lambda path: dataframe)
    monkeypatch.setattr(fci_module.utils, "graph_from_dot_file", lambda path: "ref_graph")

    def fake_fit_predict(self, X, X_test=None, ref_graph=None):
        if with_dag:
            self.dag = nx.DiGraph()
            self.dag.add_edge("x", "y")
        else:
            class DummyPag:
                def __init__(self):
                    self._edges = [("y", "x")]

                def edges(self):
                    return list(self._edges)

            self.dag = None
            self.pag = DummyPag()
        self.metrics = {"metric": 1} if metrics_available else None
        return self

    monkeypatch.setattr(fci_module.FCI, "fit_predict", fake_fit_predict)

    fci_module.main(dataset_name, input_path="input/", output_path="output/")

    output_lines = capsys.readouterr().out.splitlines()

    if with_dag:
        assert "('x', 'y')" in output_lines
    else:
        assert "('y', 'x')" in output_lines

    if metrics_available:
        assert "{'metric': 1}" in output_lines
    else:
        assert "No metrics available" in output_lines



import numpy as np

from causalexplain.estimators.cam import selGamBoost as sel_gam_boost_module


def test_sel_gamboost_returns_mask(monkeypatch):
    monkeypatch.setattr(
        sel_gam_boost_module,
        "train_GAMboost",
        lambda X, y, pars=None: {"model": type("M", (), {"feature_importances_": np.array([0.5, 0.2])})()},
    )

    X = np.array([[1.0, 2.0, 3.0], [1.5, 2.5, 3.5]])
    selection = sel_gam_boost_module.selGamBoost(X, pars={"atLeastThatMuchSelected": 0.1, "atMostThatManyNeighbors": 2}, k=1)

    assert len(selection) == X.shape[1]
    assert selection == [False, False, False]


def test_sel_gamboost_single_variable():
    X = np.ones((2, 1))
    assert sel_gam_boost_module.selGamBoost(X) == []



import numpy as np
import pytest

from causalexplain.generators import mechanisms


@pytest.fixture(autouse=True)
def _seed():
    np.random.seed(0)


def test_gmm_cause_range():
    samples = mechanisms.gmm_cause(10, k=3, verbose=True)
    assert len(samples) == 10
    assert ((samples >= -1) & (samples <= 1)).all()


def test_linear_mechanism_zero_parents():
    mech = mechanisms.LinearMechanism(ncauses=0, points=4)
    output = mech(np.empty((4, 0)))
    assert output.shape == (4, 1)


def test_polynomial_mechanism_combines_causes():
    mech = mechanisms.Polynomial_Mechanism(ncauses=1, points=6, verbose=True)
    causes = np.ones((6, 1))
    result = mech(causes, verbose=True)
    assert result.shape == (6, 1)
    assert not np.allclose(result, 0)


def test_sigmoid_mechanisms_shapes():
    causes = np.ones((5, 2))
    sig_add = mechanisms.SigmoidAM_Mechanism(2, 5)
    sig_mix = mechanisms.SigmoidMix_Mechanism(2, 5)
    assert sig_add(causes).shape == (5, 1)
    assert sig_mix(causes).shape == (5, 1)


def test_gaussian_process_add_transitions():
    """Repeated calls move the GP mechanism across its initialization branches."""
    gp = mechanisms.GaussianProcessAdd_Mechanism(ncauses=1, points=8)
    causes = np.linspace(-1, 1, 8).reshape(8, 1)
    for _ in range(6):
        result = gp(causes)
        assert result.shape == (8, 1)


def test_gaussian_process_mix_with_and_without_causes():
    """GP mix should accept both multi-cause input and the noise-only fallback."""
    gp = mechanisms.GaussianProcessMix_Mechanism(ncauses=1, points=6)
    causes = np.linspace(-1, 1, 6).reshape(6, 1)
    first = gp(causes)
    assert first.shape == (6, 1)
    gp_empty = mechanisms.GaussianProcessMix_Mechanism(ncauses=0, points=5)
    assert gp_empty(np.empty((5, 0))).shape == (5, 1)


def test_gaussian_cause_and_noise_helpers():
    assert mechanisms.gaussian_cause(4).shape == (4,)
    assert mechanisms.noise(4, v=0.5).shape == (4, 1)


def test_compute_gauss_kernel_identity():
    x = np.zeros((3, 1))
    kernel = mechanisms.computeGaussKernel(x)
    assert np.allclose(np.diag(kernel), 1.0)
